! $Id: Direct2DTest.f90 2119 2016-05-13 02:57:51Z ian $
! Copyright 2016 M.E.G.M.S.  See LICENCE.txt for licence (Apache 2.0).
! See http://www.megms.com.au/winapi-bindings.htm for more information.

!> @file
!! Defines the Direct2DTest module (which also hosts the entry point for 
!! a Windows program).


!*******************************************************************************
!!
!> Example of using Direct2D from Fortran.

MODULE Direct2DTest
  
  USE DirectWrite, ONLY: IDWriteFactory
  USE WinImagingComponent, ONLY: IWICImagingFactory, IWICBitmapSource
  USE Direct2D
  
  IMPLICIT NONE
  
  PRIVATE
  
  PUBLIC :: WinMain           ! Just for documentation
  
  !-----------------------------------------------------------------------------
  
  !> The parent type for all test types.
  TYPE, ABSTRACT :: test_type
  CONTAINS
    !> Create the device dependent resources required by the test.
    PROCEDURE :: CreateResources => test_CreateResources
    !> Discard any device dependent resources created in create_resource.
    PROCEDURE :: DiscardResources => test_DiscardResources
    !> Do the drawing of the test.
    PROCEDURE(test_Draw), DEFERRED :: Draw
    !> Adjust test resources to a different window size.
    PROCEDURE :: resize => test_Resize
  END TYPE test_type
  
  ! Interfaces for deferred bindings in test_type.
  ABSTRACT INTERFACE
    !> Do the drawing associated with the particular test.
    SUBROUTINE test_Draw(test, render_target, hr)
      USE Direct2D
      IMPORT :: test_type
      IMPLICIT NONE
      !-------------------------------------------------------------------------
      !> The test object.
      CLASS(test_type), INTENT(INOUT) :: test
      !> The render target to draw to.
      !!
      !! The test code is responsible for calling BeginDraw/EndDraw.
      CLASS(ID2D1RenderTarget), INTENT(IN) :: render_target
      !> COM result code - negative on error.  
      !!
      !! Ensure that if EndDraw encounters D2DERR_RECREATE_TARGET that it is 
      !! passed back!.
      INTEGER(HRESULT), INTENT(OUT) :: hr
    END SUBROUTINE test_Draw
  END INTERFACE
  
  
  !> A test that simply involves clearing the window.
  TYPE, EXTENDS(test_type) :: clear_test
  CONTAINS
    PROCEDURE :: draw => clear_draw
  END TYPE clear_test
  
  
  !> A test of most of the basic render target API's.
  TYPE, EXTENDS(test_type) :: general_test
    !> A brush that is always a grey (silver actually) colour.
    TYPE(ID2D1SolidColorBrush) :: grey_brush
    !> A brush that has its colour varied constantly.
    TYPE(ID2D1SolidColorBrush) :: utility_brush
  CONTAINS
    !> Create device dependent resources.
    PROCEDURE :: CreateResources => general_CreateResources
    !> Discard device dependent resources.
    PROCEDURE :: DiscardResources => general_DiscardResources
    !> Draw the test.
    PROCEDURE :: Draw => general_Draw
  END TYPE general_test
  
  
  ! Represents an block in the colour_test diagram.
  TYPE :: colour_block
    ! The names of the colours in the block, in the order that they appear.
    CHARACTER(20), ALLOCATABLE :: names(:)
  END TYPE colour_block
  
  
  !> A test that draws the colour blocks.
  TYPE, EXTENDS(test_type) :: colour_test
    !> The colour blocks that we want to draw.  This is a "constant", which 
    !! we define every time an object is created due to the allocatable 
    !! components.  We could perhaps do something clever with length 
    !! type parameters here.
    TYPE(colour_block), ALLOCATABLE :: colour_blocks(:)
    
    !> The geometry object for the colour swatch.  This is a device 
    !! independent resource, so it lives for as long as the test object 
    !! does.
    TYPE(ID2D1PathGeometry) :: hexagon
    
    !> A brush that has its colour varied constantly.
    TYPE(ID2D1SolidColorBrush) :: brush
  CONTAINS
    !> Create device dependent resources.
    PROCEDURE :: CreateResources => colour_CreateResources
    !> Discard device dependent resources.
    PROCEDURE :: DiscardResources => colour_DiscardResources
    !> Draw the test.
    PROCEDURE :: Draw => colour_Draw
  END TYPE colour_test
  
  
  !> A test of transforms.
  TYPE, EXTENDS(test_type) :: transform_test
    !> A geometry that looks a bit like an arrow.
    CLASS(ID2D1Geometry), ALLOCATABLE :: arrow
    
    !> A brush that has its colour varied constantly.
    TYPE(ID2D1SolidColorBrush) :: brush
  CONTAINS
    !> Create device dependent resources.
    PROCEDURE :: CreateResources => transform_CreateResources
    !> Discard device dependent resources.
    PROCEDURE :: DiscardResources => transform_DiscardResources
    !> Draw the test.
    PROCEDURE :: Draw => transform_Draw
  END TYPE transform_test
  
  
  !> A test of clipping.
  TYPE, EXTENDS(test_type) :: clip_test
    !> Geometry used for clipping.
    TYPE(ID2D1PathGeometry) :: hexagon
    
    !> A rectangle geometry for clipping operations.
    TYPE(ID2D1RectangleGeometry) :: rectangle_geometry
    
    !> A brush that has its colour varied constantly.
    TYPE(ID2D1SolidColorBrush) :: brush
    
    !> A layer for clipping operations.
    TYPE(ID2D1Layer) :: layer
  CONTAINS
    !> Create device dependent resources.
    PROCEDURE :: CreateResources => clip_CreateResources
    !> Discard device dependent resources.
    PROCEDURE :: DiscardResources => clip_DiscardResources
    !> Draw the test.
    PROCEDURE :: Draw => clip_Draw
  END TYPE clip_test
  
  
  !> Information for a particular line segment.
  TYPE line_data
    !> Start point of the line.
    TYPE(D2D1_POINT_2F) :: start
    !> Finish point of the line.
    TYPE(D2D1_POINT_2F) :: finish
    !> Colour of the line.
    TYPE(D2D1_COLOR_F) :: colour
    !> Width of the line.
    REAL(FLOAT) :: width
  END TYPE line_data
  
  !> A test of drawing lots of lines.
  TYPE, EXTENDS(test_type) :: line_test
    !> The data for the lines.
    TYPE(line_data), ALLOCATABLE :: lines(:)
    
    !> A brush that has its colour varied constantly.
    TYPE(ID2D1SolidColorBrush) :: brush
  CONTAINS
    !> Create device dependent resources.
    PROCEDURE :: CreateResources => line_CreateResources
    !> Discard device dependent resources.
    PROCEDURE :: DiscardResources => line_DiscardResources
    !> Draw the test.
    PROCEDURE :: Draw => line_Draw
  END TYPE line_test
  
  
  !> Information for a particular ellipse
  TYPE ellipse_data
    !> The geometry of the ellipse.
    TYPE(D2D1_ELLIPSE) :: ellipse
    !> Colour of the border.
    TYPE(D2D1_COLOR_F) :: border_colour
    !> Width of the border.
    REAL(FLOAT) :: width
    !> Colour of the fill.
    TYPE(D2D1_COLOR_F) :: fill_colour
  END TYPE ellipse_data
  
  !> A test of drawing lots of ellipses. 
  TYPE, EXTENDS(test_type) :: ellipse_test
    !> The data for the lines.
    TYPE(ellipse_data), ALLOCATABLE :: ellipses(:)
    
    !> A brush that has its colour varied constantly.
    TYPE(ID2D1SolidColorBrush) :: brush
  CONTAINS
    !> Create device dependent resources.
    PROCEDURE :: CreateResources => ellipse_CreateResources
    !> Discard device dependent resources.
    PROCEDURE :: DiscardResources => ellipse_DiscardResources
    !> Draw the test.
    PROCEDURE :: Draw => ellipse_Draw
  END TYPE ellipse_test
  
  
  !> A test of bitmaps and bitmap brushes.
  TYPE, EXTENDS(test_type) :: bmpbrush_test
    !> The shape to fill with the bitmap brush.
    TYPE(ID2D1PathGeometry) :: fill_shape
    !> Factory for Windows Imaging Components.
    TYPE(IWICImagingFactory) :: wic_factory
    !> The test bitmap, converted into a Direct2D compatible form.
    CLASS(IWICBitmapSource), ALLOCATABLE :: wic_bitmap_source
    !> A brush based on a bitmap drawn in CreateResources.
    TYPE(ID2D1BitmapBrush) :: bitmap_brush
    !> A solid brush for outlines.
    TYPE(ID2D1SolidColorBrush) :: solid_brush
    !> A d2d bitmap created from wic_btimap_source.
    TYPE(ID2D1Bitmap) :: chook
    !> A brush created from wic_bitmap_source.
    TYPE(ID2D1BitmapBrush) :: chook_brush
  CONTAINS
    FINAL :: bmpbrush_final_suspect_ifort_bug
    !> Create device dependent resources.
    PROCEDURE :: CreateResources => bmpbrush_CreateResources
    !> Discard device dependent resources.
    PROCEDURE :: DiscardResources => bmpbrush_DiscardResources
    !> Draw the test.
    PROCEDURE :: Draw => bmpbrush_Draw
  END TYPE bmpbrush_test
  
  
  !> Gratuitous advertising.
  TYPE, EXTENDS(test_type) :: logo_test
    !> The shape of the upper seam.
    TYPE(ID2D1PathGeometry) :: upper_seam
    !> The shape of the lower seam.
    TYPE(ID2D1PathGeometry) :: lower_seam
    !> The shape of the top right ellipse.
    TYPE(ID2D1TransformedGeometry) :: ellipse
    !> The shape of the letters.
    TYPE(ID2D1PathGeometry) :: letters
    !> The terracotta brush.
    TYPE(ID2D1SolidColorBrush) :: waste
    !> The black brush.
    TYPE(ID2D1SolidColorBrush) :: coal
    !> The white brush.
    TYPE(ID2D1SolidColorBrush) :: hole
  CONTAINS
    ! See comments immediately before the procedure definition.  If this 
    ! final procedure is disabled then we get a Direct2D resource leak.
    FINAL :: logo_final_suspect_ifort_bug
    !> Create device dependent resources.
    PROCEDURE :: CreateResources => logo_CreateResources
    !> Discard device dependent resources.
    PROCEDURE :: DiscardResources => logo_DiscardResources
    !> Draw the test.
    PROCEDURE :: Draw => logo_Draw
  END TYPE logo_test
  
  !-----------------------------------------------------------------------------
  
  !> The factory for producing the Direct2D resources.
  TYPE(ID2D1Factory) :: factory
  
  !> The render target for drawing to the window.  One render target - 
  !! one window.
  TYPE(ID2D1HwndRenderTarget) :: render_target
  
  !> A factory for creating things that let us use another thing created by 
  !! another factory for drawing text.  You just cannot have too many 
  !! factories.
  TYPE(IDWriteFactory) :: write_factory
  
  !> How many tests have we written.  This is doomed to be small...
  INTEGER, PARAMETER :: num_tests = 8
  
  !> Number of the current test under execution.
  INTEGER :: test_code
  
  !> The object for the current test.
  CLASS(test_type), ALLOCATABLE :: current_test
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Test entry point.
  
  FUNCTION WinMain(hInstance, hPrevInstance, lpCmdLine, nCmdShow)
    
    USE COMBase, ONLY: Exists, Release, CoInitialize, CoUninitialize
    USE Direct2D
    USE DirectWrite
    USE IFWin
    
    USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_INT, C_INTPTR_T,  &
        C_NULL_CHAR, C_PTR, C_F_POINTER, C_LOC
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The instance handle for this program.
    INTEGER(HANDLE), INTENT(IN) :: hInstance
    !> The instance handle for this program when it was 16 bit.
    INTEGER(HANDLE), INTENT(IN) :: hPrevInstance
    !> Our command line.
    CHARACTER, INTENT(IN), TARGET :: lpCmdLine(*)
    !> Window show command.
    INTEGER(C_INT), INTENT(IN) :: nCmdShow
    
    !> The program result code.
    INTEGER(C_INT) :: WinMain
    !DEC$ ATTRIBUTES STDCALL :: WinMain
    !DEC$ ATTRIBUTES DECORATE :: WinMain
    !DEC$ ATTRIBUTES ALIAS:'WinMain' :: WinMain
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Window class structure used to register our classy window.
    TYPE(T_WNDCLASS) :: wndclass
    ! The function result of registering our window class.
    INTEGER(WORD) :: class_atom
    ! The name of our window class.
    CHARACTER(:), ALLOCATABLE :: class_name
    
    ! Handle to the main window.
    INTEGER(HANDLE) :: hwnd
    ! Window title.
    CHARACTER(:), ALLOCATABLE :: window_name
    
    ! API result code.
    INTEGER(BOOL) :: brc
    ! Windows message structure.
    TYPE(T_MSG) :: msg
    ! API result code.
    INTEGER(LONG_PTR) :: lrc
    
    ! COM related function call result code.
    INTEGER(HRESULT) :: hr
    
    !***************************************************************************
    
    WinMain = 0
    
    !---------------------------------------------------------------------------
    ! Work out the initial test to display
    
    test_code = 1     ! By default.
    BLOCK
      INTEGER :: length
      
      ! Length of null terminated command line.
      length = 0
      DO
        IF (lpCmdLine(length+1) == C_NULL_CHAR) EXIT
        length = length + 1
      END DO
      
      BLOCK
        CHARACTER(length), POINTER :: cmdline
        CHARACTER(10) :: fmt
        CALL C_F_POINTER(C_LOC(lpCmdLine), cmdline)
        
        IF (LEN_TRIM(cmdline) /= 0) THEN
          ! Convert the string into an integer.  This isn't particularly 
          ! robust, but we don't care - this is just a test!
          WRITE (fmt, "('(I',I0,')')") LEN_TRIM(cmdline)
          READ (cmdline, fmt) test_code
          IF (test_code < 0 .OR. test_code > num_tests) test_code = 0
        END IF
      END BLOCK
    END BLOCK
    
    !---------------------------------------------------------------------------
    ! Create a D2D factory and similar.
    
    ! Initialize COM.  This is only required by the Windows Imaging component, 
    ! Direct2D and DirectWrite seem to get along fine without it.
    CALL CoInitialize(hr)
    
    ! We enable debugging - if the test is run under a debugger then the 
    ! debugging layer will report any issues to the debugger console.
    CALL D2D1CreateFactory( &
        D2D1_FACTORY_TYPE_SINGLE_THREADED,  &
        factory,  &
        hr,  &
        D2D1_FACTORY_OPTIONS(D2D1_DEBUG_LEVEL_WARNING) )
    IF (hr /= 0) RETURN
    
    CALL DWriteCreateFactory(DWRITE_FACTORY_TYPE_SHARED, write_factory, hr)
    IF (hr /= 0) RETURN
    
    !---------------------------------------------------------------------------
    ! Register window class.
    
    CALL ZeroMemory(LOC(wndclass), SIZEOF(wndclass))
    wndclass%style = 0
    wndclass%lpfnWndProc = LOC(WndProc)
    wndclass%cbClsExtra = 0
    wndclass%cbWndExtra = 0
    wndclass%hInstance = 0
    wndclass%hIcon = 0
    wndclass%hCursor = LoadCursor(NULL, IDC_CROSS)
    wndclass%hbrBackground = COLOR_WINDOW + 1
    wndclass%lpszMenuName = 0
    class_name = 'Direct2DRender' // C_NULL_CHAR
    wndclass%lpszClassName = LOC(class_name)
    
    class_atom = RegisterClass(wndclass)
    IF (class_atom == 0_WORD) THEN
      ! Actually returns an int, but we don't care.
      brc = MessageBox( &
          0_HANDLE,  &
          'Window class creation failed!' // C_NULL_CHAR,  &
          'Direct2D test' // C_NULL_CHAR,  &
          MB_ICONERROR )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Create a window.
    
    window_name = 'Direct2D test - press n and p to move between tests'  &
        // C_NULL_CHAR
    hwnd = CreateWindow(  &
        class_name,  &
        window_name,  &
        IANY([  &
          WS_SIZEBOX,  &
          WS_CAPTION,  &
          WS_SYSMENU,  &
          WS_MAXIMIZEBOX,  &
          WS_MINIMIZEBOX ]),  &
        100_C_INT, 100_C_INT,  &
        500_C_INT, 500_C_INT,  &
        0_HANDLE,  &
        0_HANDLE,  &
        0_HANDLE,  &
        0_LPVOID )
    IF (hwnd == 0_HANDLE) THEN
      ! Actually returns an int, but we don't care.
      brc = MessageBox( &
          0_HANDLE,  &
          'Main window creation failed!' // C_NULL_CHAR,  &
          'Direct2D test' // C_NULL_CHAR,  &
          MB_ICONERROR )
      RETURN
    END IF
    brc = ShowWindow(hwnd, nCmdShow)
    
    ! Message loop.
    DO
      brc = GetMessage(msg, 0_HANDLE, 0_UINT, 0_UINT)
      IF (brc == 0) THEN
        WinMain = msg%wparam
        EXIT
      END IF
      IF (brc == -1) THEN
        WinMain = 0
        EXIT
      END IF
      brc = TranslateMessage(msg)
      lrc = DispatchMessage(msg)
    END DO
    
    !---------------------------------------------------------------------------
    
    IF (Exists(render_target)) THEN
      IF (ALLOCATED(current_test)) THEN
        CALL current_test%DiscardResources
      END IF
      CALL Release(render_target)
    END IF
    ! We explicitly deallocate (and hence finalize) the test before we 
    ! deallocate the factories, to ensure that resources created by the 
    ! factories are released prior to the factories being released.
    IF (ALLOCATED(current_test)) DEALLOCATE(current_test)
    IF (Exists(factory)) CALL Release(factory)
    IF (Exists(write_factory)) CALL Release(write_factory)
    
    CALL CoUninitialize(hr)
    
  END FUNCTION WinMain
  
  
  !*****************************************************************************
  !!
  !> The window procedure for the test application.
  
  RECURSIVE FUNCTION WndProc(hwnd, uMsg, wParam, lParam)
    
    USE COMBase, ONLY: Exists, Release
    USE Direct2D
    USE IFWin
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The window handle addressed by the message.
    INTEGER(HANDLE), INTENT(IN) :: hwnd
    !> The message number.
    INTEGER(UINT), INTENT(IN) :: uMsg
    !> Message parameter.
    INTEGER(UINT_PTR), INTENT(IN) :: wParam
    !> Message parameter.
    INTEGER(LONG_PTR), INTENT(IN) :: lParam
    
    !> Window procedure result code.
    INTEGER(LONG_PTR) :: WndProc
    !DEC$ ATTRIBUTES STDCALL :: WndProc
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER(BOOL) :: brc      ! BOOL api result code.
    INTEGER(HRESULT) :: hr    ! COM result code.
    
    !***************************************************************************
    
    WndProc = 0   ! Default value unless specific processing decides otherwise.
    
    msg_select: SELECT CASE (uMsg)
    CASE (WM_CREATE) msg_select
      CALL allocate_test(test_code, current_test)
      
    CASE (WM_DESTROY) msg_select
      ! Closing the main window terminates the message loop.
      CALL PostQuitMessage(0)
      
    CASE (WM_PAINT) msg_select
      CALL handle_wm_paint(hwnd)
      
    CASE (WM_SIZE) msg_select
      IF (Exists(render_target)) THEN
        ! Let the render target know.
        BLOCK
          TYPE(T_RECT) :: window_rect
          
          brc = GetClientRect(hwnd, window_rect)
          CALL Resize(  &
              render_target,  &
              D2D1_SIZE_U( &
                window_rect%right - window_rect%left,  &
                window_rect%bottom - window_rect%top ),  &
              hr )
          IF (ALLOCATED(current_test)) THEN
            CALL current_test%Resize(render_target, hr)
          END IF
          IF (hr < 0) RETURN
        END BLOCK
      END IF
      brc = InvalidateRect(hwnd, NULL, FALSE)
      
    CASE (WM_DISPLAYCHANGE) msg_select
      brc = InvalidateRect(hwnd, NULL, FALSE)
      
    CASE (WM_CHAR) msg_select
      wm_char_block: BLOCK
        INTEGER :: old_test_code
        CLASS(test_type), ALLOCATABLE :: old_test
        
        old_test_code = test_code
        SELECT CASE (wParam)
        CASE (TRANSFER('n', VK_ESCAPE))
          test_code = test_code + 1
          IF (test_code > num_tests) test_code = num_tests
        CASE (TRANSFER('p', VK_ESCAPE))
          test_code = test_code - 1
          IF (test_code < 0) test_code = 0
        CASE (TRANSFER(' ', VK_ESCAPE))     ! Just redraw
          brc = InvalidateRect(hwnd, NULL, FALSE)
          EXIT msg_select
        CASE DEFAULT
          brc = MessageBeep(MB_ICONINFORMATION)
          EXIT msg_select
        END SELECT
        
        IF (old_test_code /= test_code) THEN
          IF (Exists(render_target) .AND. ALLOCATED(current_test)) THEN
            CALL current_test%DiscardResources
          END IF
          ! Work around for ifort finalization bug.
          IF (ALLOCATED(current_test)) DEALLOCATE(current_test)
          CALL allocate_test(test_code, current_test)
          IF (Exists(render_target) .AND. ALLOCATED(current_test)) THEN
            CALL current_test%CreateResources(render_target, hr)
          END IF
          brc = InvalidateRect(hwnd, NULL, FALSE)
        END IF
      END BLOCK wm_char_block
      
    CASE DEFAULT msg_select
      WndProc = DefWindowProc(hwnd, uMsg, wParam, lParam)
      
    END SELECT msg_select
    
  END FUNCTION WndProc
  
  
  !*****************************************************************************
  !!
  !> WM_PAINT processing
  
  SUBROUTINE handle_wm_paint(hwnd)
    
    USE COMBase, ONLY: Exists, Release
    USE Direct2D
    USE IFWIN
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The handle of the window being repainted.
    INTEGER(HANDLE), INTENT(IN) :: hwnd
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER(BOOL) :: brc      ! API result code.
    INTEGER(HRESULT) :: hr    ! COM result code.
    INTEGER(HANDLE) :: hdc    ! DC for painting.
    
    ! Paint struct from BeginPaint.
    TYPE(T_PAINTSTRUCT) :: ps
    
    !***************************************************************************
    
    ! Create device dependent resources if they have been discarded 
    ! (or are yet to be created) in the meantime.  If this fails, then 
    ! we call ValidateRect to stop the window getting continuously bombarded 
    ! with paint messages.
    IF (.NOT. EXISTS(render_target)) THEN
      BLOCK
        ! Window rectangle.
        TYPE(T_RECT) :: rect
        ! Window size.
        TYPE(D2D1_SIZE_U) :: size
          
        brc = GetClientRect(hwnd, rect)
      
        size = D2D1_SIZE_U(rect%right - rect%left, rect%bottom - rect%top)
        CALL CreateHwndRenderTarget(  &
            factory,  &
            D2D1_RENDER_TARGET_PROPERTIES(),  &
            D2D1_HWND_RENDER_TARGET_PROPERTIES(hwnd, size),  &
            render_target,  &
            hr )
        IF (hr < 0) THEN
          brc = ValidateRect(hwnd, 0)
          RETURN
        END IF
      END BLOCK
      
      IF (ALLOCATED(current_test)) THEN
        CALL current_test%CreateResources(render_target, hr)
        IF (hr < 0) THEN
          brc = ValidateRect(hwnd, 0)
          RETURN
        END IF
      END IF
    END IF
    
    ! This apparently (from testing) causes the render target to be clipped 
    ! to the update rectangle, though I haven't found that explicitly 
    ! documented.  An alternative to a BeginPaint/EndPaint pair 
    ! is to call `ValidateRect(hwnd, 0)`
    hdc = BeginPaint(hwnd, ps)
    
    IF (ALLOCATED(current_test)) THEN
      ! When tests fail, they typically just stop drawing, or maybe they 
      ! crash - we don't really care as long as there is some visual 
      ! impact, but perhaps we could be a bit more thorough about reporting 
      ! unexpected API failures.
      CALL current_test%Draw(render_target, hr)
      IF (hr == D2DERR_RECREATE_TARGET) THEN
        CALL current_test%DiscardResources
        IF (Exists(render_target)) CALL Release(render_target)
      ELSE IF (hr < 0) THEN
        BLOCK
          CHARACTER(:), ALLOCATABLE :: msg
          
          msg = 'Test drawing failed'
          brc = TextOut(hdc, 10, 10, msg, LEN(msg))
        END BLOCK
      END IF
    ELSE
      BLOCK
        CHARACTER(:), ALLOCATABLE :: msg
        
        brc = FillRect(hdc, ps%rcPaint, GetSysColorBrush(COLOR_WINDOW))
        msg = 'Test construction failed'
        brc = TextOut(hdc, 10, 10, msg, LEN(msg))
      END BLOCK
    END IF
    
    brc = EndPaint(hdc, ps)
    
  END SUBROUTINE handle_wm_paint
  
  
  !*****************************************************************************
  !!
  !> Creates the relevant test object for the given test number.
  
  SUBROUTINE allocate_test(number, test)
    
    USE Direct2D, ONLY: HRESULT
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test number.
    INTEGER, INTENT(IN) :: number
    !> The test object.  Not allocated on error.
    CLASS(test_type), INTENT(OUT), ALLOCATABLE :: test
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER(HRESULT) :: hr    ! COM result - negative on error.
    
    !***************************************************************************
    
    ! ifort does not yet support assignment to a polymorphic object.  The 
    ! usual F2003 work around of using ALLOCATE(xxx, SOURCE=yyy) does not 
    ! work here, because we need defined assignment to operate to ensure 
    ! that COM reference counting is handled correctly (ALLOCATE does 
    ! not invoke defined assignment, but the allocated object is still 
    ! finalized when it is deallocated).
    !
    ! So instead we use MOVE_ALLOC to move a non-polymorphic temporary 
    ! object into our abstract polymorphic result.
    !
    ! Initially we used functions for constructors, but it is difficult 
    ! to report failure from a function, so now we use subroutines.
    
    SELECT CASE (number)
    CASE DEFAULT
      BLOCK
        TYPE(clear_test), ALLOCATABLE :: test_tmp
        ALLOCATE(test_tmp)
        CALL construct_clear_test(test_tmp, hr)
        IF (hr < 0) RETURN
        CALL MOVE_ALLOC(test_tmp, test)
      END BLOCK
      
    CASE (1)
      BLOCK
        TYPE(general_test), ALLOCATABLE :: test_tmp
        ALLOCATE(test_tmp)
        CALL construct_general_test(test_tmp, hr)
        IF (hr < 0) RETURN
        CALL MOVE_ALLOC(test_tmp, test)
      END BLOCK
      
    CASE (2)
      BLOCK
        TYPE(colour_test), ALLOCATABLE :: test_tmp
        ALLOCATE(test_tmp)
        CALL construct_colour_test(test_tmp, hr)
        IF (hr < 0) RETURN
        CALL MOVE_ALLOC(test_tmp, test)
      END BLOCK
      
    CASE (3)
      BLOCK
        TYPE(transform_test), ALLOCATABLE :: test_tmp
        ALLOCATE(test_tmp)
        CALL construct_transform_test(test_tmp, hr)
        IF (hr < 0) RETURN
        CALL MOVE_ALLOC(test_tmp, test)
      END BLOCK
      
    CASE (4)
      BLOCK
        TYPE(clip_test), ALLOCATABLE :: test_tmp
        ALLOCATE(test_tmp)
        CALL construct_clip_test(test_tmp, hr)
        IF (hr < 0) RETURN
        CALL MOVE_ALLOC(test_tmp, test)
      END BLOCK
      
    CASE (5)
      BLOCK
        TYPE(line_test), ALLOCATABLE :: test_tmp
        ALLOCATE(test_tmp)
        CALL construct_line_test(test_tmp, hr)
        IF (hr < 0) RETURN
        CALL MOVE_ALLOC(test_tmp, test)
      END BLOCK
      
    CASE (6)
      BLOCK
        TYPE(ellipse_test), ALLOCATABLE :: test_tmp
        ALLOCATE(test_tmp)
        CALL construct_ellipse_test(test_tmp, hr)
        IF (hr < 0) RETURN
        CALL MOVE_ALLOC(test_tmp, test)
      END BLOCK
      
    CASE (7)
      BLOCK
        TYPE(bmpbrush_test), ALLOCATABLE :: test_tmp
        ALLOCATE(test_tmp)
        CALL construct_bmpbrush_test(test_tmp, hr)
        IF (hr < 0) RETURN
        CALL MOVE_ALLOC(test_tmp, test)
      END BLOCK
      
    CASE (8)
      BLOCK
        TYPE(logo_test), ALLOCATABLE :: test_tmp
        ALLOCATE(test_tmp)
        CALL construct_logo_test(test_tmp, hr)
        IF (hr < 0) RETURN
        CALL MOVE_ALLOC(test_tmp, test)
      END BLOCK
    END SELECT
    
  END SUBROUTINE allocate_test
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for test_type.
  
  
  !*****************************************************************************
  !!
  !> Default create_resources implementation
  !!
  !! Called after the render target has been created.
  !!
  !! This default implementation creates no resources.
  
  SUBROUTINE test_CreateResources(test, render_target, hr)
    
    USE Direct2D
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(test_type), INTENT(INOUT) :: test
    !> The render target.
    CLASS(ID2D1RenderTarget), INTENT(IN) :: render_target
    !> COM result code - negative on error.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !***************************************************************************
    
    hr = 0_HRESULT
    
  END SUBROUTINE test_CreateResources
  
  
  !*****************************************************************************
  !!
  !> Default discard_resources implementation
  !!
  !! Called before the render target has been discarded.
  !!
  !! This default implementation discards no resources.
  
  SUBROUTINE test_DiscardResources(test)
    
    USE COMBase
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(test_type), INTENT(INOUT) :: test
    
    !***************************************************************************
    
  END SUBROUTINE test_DiscardResources
  
  
  !*****************************************************************************
  !!
  !> Respond to a resize of the window.
  !!
  !! Called after the render target has itself been resized.  Not called 
  !! if the render target does not exist at the time the resize happens.
  !!
  !! This default implementation does nothing.
  
  SUBROUTINE test_Resize(test, render_target, hr)
    
    USE Direct2D
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(test_type), INTENT(INOUT) :: test
    !> The render target.
    CLASS(ID2D1RenderTarget), INTENT(IN) :: render_target
    !> COM result code - negative on error.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !***************************************************************************
    
    hr = 0_HRESULT
    
  END SUBROUTINE test_Resize
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for clear_test.
  
  
  !*****************************************************************************
  !!
  !> Construct a clear_test object.
  !!
  !! We have written this just to make the tests consistent.
  
  SUBROUTINE construct_clear_test(test, hr)
    
    USE Direct2D
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    !> The constructed object.
    TYPE(clear_test), INTENT(OUT) :: test
    !> COM result code - negative on error.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !***************************************************************************
    
    ! Nothing to do.
    hr = 0_HRESULT
    
    ! This simply silences some compiler warnings.
    test = clear_test()
    
  END SUBROUTINE construct_clear_test
  
  
  !*****************************************************************************
  !!
  !> A test of a nice clear screen.
  
  SUBROUTINE clear_Draw(test, render_target, hr)
    
    USE Direct2D
    USE Direct2DColor
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(clear_test), INTENT(INOUT) :: test
    !> The render target to draw to.
    CLASS(ID2D1RenderTarget), INTENT(IN) :: render_target
    !> COM result code - negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !***************************************************************************
    
    CALL BeginDraw(render_target)
    
    CALL SetTransform(render_target, IdentityMatrix)
    CALL Clear(render_target, LightSalmon)
    
    CALL EndDraw(render_target, hr)
    
  END SUBROUTINE clear_Draw
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for general_test.
  
  
  !*****************************************************************************
  !!
  !> Construct a general_test object.
  !!
  !! Some of the device indepedent resources could be created here rather 
  !! than general_Draw.
  
  SUBROUTINE construct_general_test(test, hr)
    
    USE Direct2D
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    !> The constructed object.
    TYPE(general_test), INTENT(OUT) :: test
    !> COM result code - negative on error.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !***************************************************************************
    
    ! Nothing to do.
    hr = 0_HRESULT
    
  END SUBROUTINE construct_general_test
  
  
  !*****************************************************************************
  !!
  !> Create the device dependent resources for the general tests.
  !!
  !! Some device dependent (and independent) resources for the general tests 
  !! are created when the test is being drawn.  This is not particularly 
  !! efficient, but because this is just a test, we don't care.
  
  SUBROUTINE general_CreateResources(test, render_target, hr)
    
    USE COMBase
    USE Direct2D
    USE Direct2DColor
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(general_test), INTENT(INOUT) :: test
    !> The current render target.
    CLASS(ID2D1RenderTarget), INTENT(IN) :: render_target
    !> COM result code - negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !***************************************************************************
    
    CALL CreateSolidColorBrush(  &
        render_target,  &
        Silver,  &
        test%grey_brush,  &
        hr )
    IF (hr < 0) RETURN
    
    CALL CreateSolidColorBrush(  &
        render_target,  &
        Black,  &
        test%utility_brush,  &
        hr )
    IF (hr < 0) RETURN
    
    BLOCK
      TYPE(ID2D1Brush) :: b2
      
      CALL QueryInterface(test%utility_brush, b2)
    END BLOCK
    
  END SUBROUTINE general_CreateResources
  
  
  !*****************************************************************************
  !!
  !> Discard the device dependent resources for the general tests.
  
  SUBROUTINE general_DiscardResources(test)
    
    USE COMBase
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(general_test), INTENT(INOUT) :: test
    
    !***************************************************************************
    
    IF (Exists(test%utility_brush))   CALL Release(test%utility_brush)
    IF (Exists(test%grey_brush))      CALL Release(test%grey_brush)
    
  END SUBROUTINE general_DiscardResources
  
  
  !*****************************************************************************
  !!
  !> A test of a whole heap of little things.
  
  SUBROUTINE general_Draw(test, render_target, hr)
    
    USE Direct2D
    USE DirectWrite
    USE Direct2DColor
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(general_test), INTENT(INOUT) :: test
    !> The current render target.
    CLASS(ID2D1RenderTarget), INTENT(IN) :: render_target
    !> COM result code - negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    REAL(FLOAT) :: xo         ! Little test rectangle origin x.
    REAL(FLOAT) :: yo         ! Little test rectangle origin y.
    
    ! The dimensions of each litle test rectangle, in dips.
    REAL(FLOAT), PARAMETER :: dx = 100.0
    REAL(FLOAT), PARAMETER :: dy = 100.0
    
    INTEGER :: itest          ! The general test we are executing.
    
    ! Window size, in dip's.
    TYPE(D2D1_SIZE_F) :: window_size
    
    ! The rectangle for the current test, in dips.
    TYPE(D2D1_RECT_F) :: test_rect
    
    ! Local COM result code - so we can separate sub-test failure from 
    ! the result of EndDraw.
    INTEGER(HRESULT) :: local_hr
    
    !***************************************************************************
    
    CALL BeginDraw(render_target)
    
    CALL SetTransform(render_target, IdentityMatrix)
    CALL Clear(render_target, AntiqueWhite)
    
    ! This is the window size in device independent pixels.  If the user 
    ! is running a display that does other than 96 dpi, this will be 
    ! different from the size of the window rectangle.
    window_size = GetSize(render_target)
    
    local_hr = 0_HRESULT
    xo = 0.0
    yo = 0.0
    test_loop: DO itest = 1, 18
      test_rect = D2D1_RECT_F(xo + 5.0, yo + 5.0, xo + dx - 5.0, yo + dy - 5.0)
      
      CALL execute_a_general_test(  &
          test,  &
          render_target,  &
          itest,  &
          test_rect,  &
          local_hr )
      IF (local_hr < 0) EXIT test_loop
      
      xo = xo + dx
      IF (xo + dx > window_size%width) THEN
        yo = yo + dy
        xo = 0.0
      END IF
    END DO test_loop
    
    CALL EndDraw(render_target, hr)
    ! If EndDraw returned some sort of status code - then return that.  
    ! Otherwise return whatever status code we got out of the 
    ! individual tests.
    IF (hr == 0) hr = local_hr
    
  END SUBROUTINE general_Draw
  
  
  !*****************************************************************************
  !!
  !> All the little general tests.
  
  SUBROUTINE execute_a_general_test(test, render_target, itest, test_rect, hr)
    
    USE Direct2D
    USE DirectWrite
    USE Direct2DColor
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(general_test), INTENT(INOUT) :: test
    !> The render target to draw to.
    CLASS(ID2D1RenderTarget), INTENT(IN) :: render_target
    !> The number of the general test to execute.
    INTEGER, INTENT(IN) :: itest
    !> The rectangle for the output of the general test.
    TYPE(D2D1_RECT_F), INTENT(IN) :: test_rect
    !> COM result - negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !***************************************************************************
    
    hr = 0_HRESULT
    
    SELECT CASE (itest)
    CASE (1)  ! DrawLine
      CALL DrawLine(  &
          render_target,  &
          D2D1_POINT_2F(test_rect%left, test_rect%top),  &
          D2D1_POINT_2F(test_rect%right, test_rect%bottom),  &
          test%grey_brush,  &
          5.0 )
      
    CASE (2)  ! DrawRectangle
      CALL DrawRectangle(  &
          render_target,  &
          test_rect,  &
          test%grey_brush )
      
    CASE (3)  ! FillRectangle
      CALL SetColor(test%utility_brush, Blue)
      CALL FillRectangle(  &
          render_target,  &
          test_rect,  &
          test%utility_brush )
      
    CASE (4)  ! DrawRoundedRectangle
      CALL SetColor(test%utility_brush, Pink)
      CALL DrawRoundedRectangle(  &
          render_target,  &
          D2D1_ROUNDED_RECT(test_rect, 15.0, 10.0),  &
          test%utility_brush )
      
    CASE (5)  ! FillRoundedRectangle
      CALL SetColor(test%utility_brush, Fuchsia)
      CALL FillRoundedRectangle(  &
          render_target,  &
          D2D1_ROUNDED_RECT(test_rect, 15.0, 10.0),  &
          test%utility_brush )
      
    CASE (6)  ! DrawEllipse
      CALL SetColor(test%utility_brush, DarkOliveGreen)
      CALL DrawEllipse(  &
          render_target,  &
          D2D1_ELLIPSE(  &
            D2D1_POINT_2F( &
              (test_rect%left + test_rect%right) / 2.0,  &
              (test_rect%top + test_rect%bottom) / 2.0 ),  &
            20.0_FLOAT,  &
            10.0_FLOAT ),  &
          test%utility_brush )
      
    CASE (7)  ! FillEllipse
      CALL SetColor(test%utility_brush, DarkSalmon)
      CALL FillEllipse(  &
          render_target,  &
          D2D1_ELLIPSE(  &
            D2D1_POINT_2F( &
              (test_rect%left + test_rect%right) / 2.0,  &
              (test_rect%top + test_rect%bottom) / 2.0 ),  &
            20.0_FLOAT,  &
            10.0_FLOAT ),  &
          test%utility_brush )
      
    CASE (8)  ! Draw some text.
      BLOCK
        TYPE(IDWriteTextFormat) :: text_format
        
        CALL CreateTextFormat(  &
            write_factory,  &
            'Arial',  &
            DWRITE_FONT_WEIGHT_NORMAL,  &
            DWRITE_FONT_STYLE_NORMAL,  &
            DWRITE_FONT_STRETCH_NORMAL,  &
            10.0_FLOAT,  &
            'C',  &
            text_format,  &
            hr )
        IF (hr < 0) RETURN
        
        CALL SetColor(test%utility_brush, YellowGreen)
        CALL DrawText(  &
            render_target,  &
            'Direct2D from Fortran',  &
            text_format,  &
            test_rect,  &
            test%utility_brush )
      END BLOCK
      
    CASE (9)    ! Linear gradient brush.
      BLOCK
        TYPE(ID2D1GradientStopCollection) :: gradient_stop_collection
        TYPE(ID2D1LinearGradientBrush) :: linear_gradient_brush
          
        CALL CreateGradientStopCollection(  &
            render_target,  &
            [ D2D1_GRADIENT_STOP(0.0_FLOAT, Green),  &
              D2D1_GRADIENT_STOP(1.0_FLOAT, Red) ],  &
            gradient_stop_collection,  &
            hr )
        IF (hr < 0) RETURN
        
        CALL CreateLinearGradientBrush(  &
            render_target,  &
            D2D1_LINEAR_GRADIENT_BRUSH_PROPERTIES(  &
              D2D1_POINT_2F(test_rect.left, test_rect.top), &
              D2D1_POINT_2F(test_rect.right, test_rect.bottom) ),  &
            gradient_stop_collection,  &
            linear_gradient_brush,  &
            hr )
        IF (hr < 0) RETURN
        
        CALL FillRoundedRectangle(  &
            render_target,  &
            D2D1_ROUNDED_RECT(test_rect, 15.0_FLOAT, 10.0_FLOAT),  &
            linear_gradient_brush )
      END BLOCK
      
    CASE (10)   ! Radial gradient brush.
      BLOCK
        TYPE(ID2D1GradientStopCollection) :: gradient_stop_collection
        TYPE(ID2D1RadialGradientBrush) :: radial_gradient_brush
        
        CALL CreateGradientStopCollection( render_target,  &
            [ D2D1_GRADIENT_STOP(0.0_FLOAT, Green),  &
              D2D1_GRADIENT_STOP(1.0_FLOAT, Red) ],  &
            gradient_stop_collection,  &
            hr )
        IF (hr < 0) RETURN
        
        CALL CreateRadialGradientBrush(  &
            render_target,  &
            D2D1_RADIAL_GRADIENT_BRUSH_PROPERTIES(  &
              D2D1_POINT_2F(  &
                (test_rect.left + test_rect.right) / 2.0_FLOAT,  &
                (test_rect.top + test_rect.bottom) / 2.0_FLOAT ),  &
              RADIUSX=20.0_FLOAT, RADIUSY=30.0_FLOAT ), &
            gradient_stop_collection,  &
            radial_gradient_brush,  &
            hr )
        IF (hr < 0) RETURN
        
        CALL FillRoundedRectangle(  &
            render_target,  &
            D2D1_ROUNDED_RECT(test_rect, 15.0_FLOAT, 10.0_FLOAT),  &
            radial_gradient_brush )
      END BLOCK
      
    CASE (11)   ! DrawTextLayout (and DWriteTextLayout stuff).
      BLOCK
        USE DirectWrite     ! ifort 17.0 beta bug.  Why ifort, why WHY WHY?????
        TYPE(IDWriteTextFormat) :: text_format
        TYPE(IDWriteTextLayout) :: text_layout
        TYPE(DWRITE_TEXT_METRICS) :: text_metrics
        
        CALL CreateTextFormat(  &
            write_factory,  &
            'Arial',  &
            DWRITE_FONT_WEIGHT_NORMAL,  &
            DWRITE_FONT_STYLE_NORMAL,  &
            DWRITE_FONT_STRETCH_NORMAL,  &
            10.0_FLOAT,  &
            'C',  &
            text_format,  &
            hr )
        IF (hr < 0) RETURN
        
        CALL SetWordWrapping(text_format, DWRITE_WORD_WRAPPING_NO_WRAP, hr)
        IF (hr < 0) RETURN
        CALL CreateTextLayout(  &
            write_factory,  &
            'Direct2D from Fortran',  &
            text_format,  &
            0.0, 0.0,  &
            text_layout,  &
            hr )
        IF (hr < 0) RETURN
        
        CALL GetMetrics(text_layout, text_metrics, hr)
        IF (hr < 0) RETURN
        
        CALL SetColor(test%utility_brush, YellowGreen)
        
        CALL DrawTextLayout(  &
            render_target,  &
            D2D1_POINT_2F(test_rect%left, test_rect%top),  &
            text_layout,  &
            test%utility_brush )
        
        ! Draw a rectangle based on the metrics.
        CALL DrawRectangle(  &
            render_target,  &
            D2D1_RECT_F(  &
              test_rect%left + text_metrics%left,  &
              test_rect%top + text_metrics%top,  &
              test_rect%left + text_metrics%width,  &
              test_rect%top + text_metrics%height ),  &
            test%grey_brush )
      END BLOCK
      
    CASE (12)   ! Rectangle geometry
      BLOCK
        TYPE(ID2D1RectangleGeometry) :: rectangle_geometry
        
        CALL CreateRectangleGeometry(  &
            factory,  &
            test_rect,  &
            rectangle_geometry,  &
            hr ) 
        IF (hr < 0) RETURN
        
        CALL SetColor(test%utility_brush, OrangeRed)
        CALL FillGeometry(  &
            render_target,  &
            rectangle_geometry,  &
            test%utility_brush )
        CALL SetColor(test%utility_brush, LightSteelBlue)
        CALL DrawGeometry(  &
            render_target,  &
            rectangle_geometry,  &
            test%utility_brush,  &
            5.0 )
      END BLOCK
      
    CASE (13)   ! Rounded rectangle geometry
      BLOCK
        TYPE(ID2D1RoundedRectangleGeometry) :: rectangle_geometry
        
        CALL CreateRoundedRectangleGeometry(  &
            factory,  &
            D2D1_ROUNDED_RECT(test_rect, 10.0_FLOAT, 10.0_FLOAT),  &
            rectangle_geometry,  &
            hr ) 
        IF (hr < 0) RETURN
        
        CALL SetColor(test%utility_brush, Goldenrod)
        CALL FillGeometry(  &
            render_target,  &
            rectangle_geometry,  &
            test%utility_brush  )
        CALL SetColor(test%utility_brush, SeaGreen)
        CALL DrawGeometry(  &
            render_target,  &
            rectangle_geometry,  &
            test%utility_brush,  &
            5.0 )
      END BLOCK
      
    CASE (14)   ! Ellipse geometry.
      BLOCK
        TYPE(ID2D1EllipseGeometry) :: ellipse_geometry
        
        CALL CreateEllipseGeometry(  &
            factory,  &
            D2D1_ELLIPSE(  &
              D2D1_POINT_2F(  &
                (test_rect%left + test_rect%right) / 2.0_FLOAT,  &
                (test_rect%top + test_rect%bottom) / 2.0_FLOAT ),  &
              (test_rect%right - test_rect%left) / 2.0,  &
              (test_rect%bottom - test_rect%top) / 4.0 ),  &
            ellipse_geometry,  &
            hr ) 
        IF (hr < 0) RETURN
        
        CALL SetColor(test%utility_brush, LightGreen)
        CALL FillGeometry(  &
            render_target,  &
            ellipse_geometry,  &
            test%utility_brush  )
        CALL SetColor(test%utility_brush, MediumOrchid)
        CALL DrawGeometry(  &
            render_target,  &
            ellipse_geometry,  &
            test%utility_brush,  &
            5.0 )
      END BLOCK
      
    CASE (15)   ! Stroked path geometry.
      BLOCK
        TYPE(ID2D1PathGeometry) :: hexagon
        TYPE(ID2D1TransformedGeometry) :: transformed_geometry
        
        CALL create_hexagon(hexagon, hr)
        IF (hr < 0) RETURN
        
        CALL CreateTransformedGeometry(  &
            factory,  &
            hexagon,  &
            Scale(4.0_FLOAT, 4.0_FLOAT, D2D1_POINT_2F())  &
              * Translation( &
                (test_rect%left + test_rect%right) / 2.0_FLOAT,  &
                (test_rect%top + test_rect%bottom) / 2.0_FLOAT ),  &
            transformed_geometry,  &
            hr )
        IF (hr < 0) RETURN
        
        CALL DrawGeometry(  &
            render_target,  &
            transformed_geometry,  &
            test%grey_brush )
      END BLOCK
      
    CASE (16)   ! Filled path geometry.
      BLOCK
        TYPE(ID2D1PathGeometry) :: hexagon
        TYPE(ID2D1TransformedGeometry) :: transformed_geometry
        
        CALL create_hexagon(hexagon, hr)
        IF (hr < 0) RETURN
        
        CALL CreateTransformedGeometry(  &
            factory,  &
            hexagon,  &
            Scale(4.0_FLOAT, 4.0_FLOAT, D2D1_POINT_2F())  &
              * Translation(  &
                (test_rect%left + test_rect%right) / 2.0_FLOAT,  &
                (test_rect%top + test_rect%bottom) / 2.0_FLOAT ),  &
            transformed_geometry,  &
            hr )
        IF (hr < 0) RETURN
        
        CALL FillGeometry(  &
            render_target,  &
            transformed_geometry,  &
            test%grey_brush )
      END BLOCK
      
    CASE (17)   ! The arrow geometry
      BLOCK
        CLASS(ID2D1Geometry), ALLOCATABLE :: arrow
        TYPE(ID2D1TransformedGeometry) :: transformed_geometry
        
        CALL create_arrow(arrow, hr)
        IF (hr < 0) RETURN
        
        CALL CreateTransformedGeometry(  &
            factory,  &
            arrow,  &
            Scale(2.0_FLOAT, 2.0_FLOAT, D2D1_POINT_2F())  &
              * Translation(  &
                (test_rect%left + test_rect%right) / 2.0_FLOAT,  &
                (test_rect%top + test_rect%bottom) / 2.0_FLOAT ),  &
            transformed_geometry,  &
            hr )
        IF (hr < 0) RETURN
        
        CALL SetColor(test%utility_brush, Red)
        CALL FillGeometry(  &
            render_target,  &
            transformed_geometry,  &
            test%utility_brush )
        CALL SetColor(test%utility_brush, Black)
        CALL DrawGeometry(  &
            render_target,  &
            transformed_geometry,  &
            test%utility_brush,  &
            1.0_FLOAT )
      END BLOCK
    
    CASE (18)     ! Stroked rectangle.
      BLOCK
        TYPE(ID2D1StrokeStyle) :: stroke_solid
        TYPE(ID2D1StrokeStyle) :: stroke_dash
        TYPE(ID2D1StrokeStyle) :: stroke_dot
        TYPE(ID2D1StrokeStyle) :: stroke_dash_dot
        TYPE(ID2D1StrokeStyle) :: stroke_dash_dot_dot
        TYPE(ID2D1StrokeStyle) :: stroke_custom
        TYPE(D2D1_RECT_F) :: rect
        
        CALL CreateStrokeStyle(  &
            factory,  &
            D2D1_STROKE_STYLE_PROPERTIES(  &
              DASHSTYLE=D2D1_DASH_STYLE_SOLID ),  &
            stroke_solid,  &
            hr )
        IF (hr < 0) RETURN
        
        CALL CreateStrokeStyle(  &
            factory,  &
            D2D1_STROKE_STYLE_PROPERTIES(  &
              DASHSTYLE=D2D1_DASH_STYLE_DASH ),  &
            stroke_dash,  &
            hr )
        IF (hr < 0) RETURN
        
        ! A dashcap of something other than the default is required in order 
        ! to see any dots in the built-in patterns.
        CALL CreateStrokeStyle(  &
            factory,  &
            D2D1_STROKE_STYLE_PROPERTIES(  &
              DASHCAP=D2D1_CAP_STYLE_ROUND,  &
              DASHSTYLE=D2D1_DASH_STYLE_DOT ),  &
            stroke_dot,  &
            hr )
        IF (hr < 0) RETURN
        
        CALL CreateStrokeStyle(  &
            factory,  &
            D2D1_STROKE_STYLE_PROPERTIES(  &
              DASHCAP=D2D1_CAP_STYLE_ROUND,  &
              DASHSTYLE=D2D1_DASH_STYLE_DASH_DOT ),  &
            stroke_dash_dot,  &
            hr )
        IF (hr < 0) RETURN
        
        CALL CreateStrokeStyle(  &
            factory,  &
            D2D1_STROKE_STYLE_PROPERTIES(  &
              DASHCAP=D2D1_CAP_STYLE_ROUND,  &
              DASHSTYLE=D2D1_DASH_STYLE_DASH_DOT_DOT ),  &
            stroke_dash_dot_dot,  &
            hr )
        IF (hr < 0) RETURN
        
        CALL CreateStrokeStyle(  &
            factory,  &
            D2D1_STROKE_STYLE_PROPERTIES(  &
              DASHSTYLE=D2D1_DASH_STYLE_CUSTOM ),  &
            stroke_custom,  &
            hr,  &
            DASHES=[ 1.0_FLOAT, 2.0_FLOAT, 3.0_FLOAT, 4.0_FLOAT,  &
                     5.0_FLOAT, 6.0_FLOAT, 7.0_FLOAT, 8.0_FLOAT ] )
        IF (hr < 0) RETURN
        
        ! Custom on the outside, then work into solid.
        rect = test_rect
        CALL SetColor(test%utility_brush, Blue)
        CALL DrawRectangle(  &
            render_target,  &
            rect,  &
            test%utility_brush,  &
            2.0_FLOAT,  &
            STROKESTYLE=stroke_custom )
        
        rect%left = rect%left + 5.0_FLOAT
        rect%right = rect.right - 5.0_FLOAT
        rect%top = rect%top + 5.0_FLOAT
        rect%bottom = rect%bottom - 5.0_FLOAT
        CALL SetColor(test%utility_brush, Green)
        CALL DrawRectangle(  &
            render_target,  &
            rect,  &
            test%utility_brush,  &
            2.0_FLOAT,  &
            STROKESTYLE=stroke_dash_dot_dot )
        
        rect%left = rect%left + 5.0_FLOAT
        rect%right = rect.right - 5.0_FLOAT
        rect%top = rect%top + 5.0_FLOAT
        rect%bottom = rect%bottom - 5.0_FLOAT
        CALL SetColor(test%utility_brush, Red)
        CALL DrawRectangle(  &
            render_target,  &
            rect,  &
            test%utility_brush,  &
            2.0_FLOAT,  &
            STROKESTYLE=stroke_dash_dot )
        
        rect%left = rect%left + 5.0_FLOAT
        rect%right = rect.right - 5.0_FLOAT
        rect%top = rect%top + 5.0_FLOAT
        rect%bottom = rect%bottom - 5.0_FLOAT
        CALL SetColor(test%utility_brush, Cyan)
        CALL DrawRectangle(  &
            render_target,  &
            rect,  &
            test%utility_brush,  &
            2.0_FLOAT,  &
            STROKESTYLE=stroke_dot )
        
        rect%left = rect%left + 5.0_FLOAT
        rect%right = rect.right - 5.0_FLOAT
        rect%top = rect%top + 5.0_FLOAT
        rect%bottom = rect%bottom - 5.0_FLOAT
        CALL SetColor(test%utility_brush, Magenta)
        CALL DrawRectangle(  &
            render_target,  &
            rect,  &
            test%utility_brush,  &
            2.0_FLOAT,  &
            STROKESTYLE=stroke_dash )
        
        rect%left = rect%left + 5.0_FLOAT
        rect%right = rect.right - 5.0_FLOAT
        rect%top = rect%top + 5.0_FLOAT
        rect%bottom = rect%bottom - 5.0_FLOAT
        CALL SetColor(test%utility_brush, Black)
        CALL DrawRectangle(  &
            render_target,  &
            rect,  &
            test%utility_brush,  &
            2.0_FLOAT,  &
            STROKESTYLE=stroke_solid )
      END BLOCK
      
    END SELECT
    
  END SUBROUTINE execute_a_general_test
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for colour_test
  
  
  !*****************************************************************************
  !!
  !> Constructs a colour_test object.
  
  SUBROUTINE construct_colour_test(test, hr)
    
    USE Direct2D
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The constructed object.
    TYPE(colour_test), INTENT(OUT) :: test
    !> COM result code - negative on error.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !***************************************************************************
    
    test%colour_blocks = [  &
        colour_block( [  &
          'White               ',  &
          'Gainsboro           ',  &
          'LightGray           ',  &
          'Silver              ',  &
          'DarkGrey            ',  &
          'Gray                ',  &
          'DimGrey             ',  &
          'Black               ' ] ),  &
        colour_block( [  &
          'GreenYellow         ',  &
          'LightGreen          ',  &
          'PaleGreen           ',  &
          'YellowGreen         ',  &
          'Chartreuse          ',  &
          'LawnGreen           ',  &
          'Lime                ',  &
          'LimeGreen           ',  &
          'Green               ',  &
          'ForestGreen         ',  &
          'DarkGreen           ',  &
          'MediumSpringGreen   ',  &
          'SpringGreen         ',  &
          'MediumSeaGreen      ',  &
          'SeaGreen            ',  &
          'DarkSeaGreen        ',  &
          'OliveDrab           ',  &
          'Olive               ',  &
          'DarkOliveGreen      ' ] ),  &
        colour_block( [  &
          'Snow                ',  &
          'FloralWhite         ',  &
          'Ivory               ',  &
          'SeaShell            ',  &
          'OldLace             ',  &
          'Cornsilk            ',  &
          'GhostWhite          ',  &
          'WhiteSmoke          ',  &
          'MintCream           ',  &
          'Azure               ',  &
          'Honeydew            ',  &
          'AliceBlue           ',  &
          'LightCyan           ',  &
          'Linen               ',  &
          'AntiqueWhite        ',  &
          'Beige               ',  &
          'BlanchedAlmond      ',  &
          'Bisque              ',  &
          'PeachPuff           ',  &
          'LavenderBlush       ',  &
          'Lavender            '] ),  &
        colour_block( [  &
          'CadetBlue           ',  &
          'DarkCyan            ',  &
          'Teal                ',  &
          'Turquoise           ',  &
          'LightSeaGreen       ',  &
          'DarkTurquoise       ',  &
          'MediumAquamarine    ',  &
          'Aquamarine          ',  &
          'Aqua                ',  &
          'Cyan                ',  &
          'MediumTurquoise     ',  &
          'PaleTurquoise       ',  &
          'PowderBlue          ',  &
          'LightBlue           ',  &
          'LightSkyBlue        ',  &
          'SkyBlue             ',  &
          'DeepSkyBlue         ',  &
          'LightSteelBlue      ',  &
          'CornflowerBlue      ',  &
          'DodgerBlue          ',  &
          'SteelBlue           ',  &
          'RoyalBlue           ',  &
          'Blue                ',  &
          'MediumBlue          ',  &
          'DarkBlue            ',  &
          'MidnightBlue        ',  &
          'Navy                ',  &
          'LightSlateGrey      ',  &
          'DarkSlateGrey       ',  &
          'SlateGrey           ',  &
          'DarkSlateBlue       ',  &
          'MediumSlateBlue     ',  &
          'SlateBlue           ',  &
          'Indigo              ',  &
          'BlueViolet          ',  &
          'MediumPurple        ',  &
          'DarkViolet          ',  &
          'DarkOrchid          ',  &
          'DarkMagenta         ',  &
          'Purple              ',  &
          'MediumOrchid        ',  &
          'Magenta             ',  &
          'Fuchsia             ',  &
          'Orchid              ',  &
          'Violet              ',  &
          'Plum                ',  &
          'Thistle             ',  &
          'PaleVioletRed       ',  &
          'MediumVioletRed     ',  &
          'DeepPink            ',  &
          'HotPink             ',  &
          'LightPink           ',  &
          'Pink                ',  &
          'MistyRose           ' ] ),  &
        colour_block( [  &
          'RosyBrown           ',  &
          'Salmon              ',  &
          'IndianRed           ',  &
          'LightCoral          ',  &
          'DarkSalmon          ',  &
          'LightSalmon         ',  &
          'SandyBrown          ',  &
          'Orange              ',  &
          'DarkOrange          ',  &
          'Coral               ',  &
          'Tomato              ',  &
          'OrangeRed           ',  &
          'Red                 ',  &
          'Crimson             ',  &
          'Firebrick           ',  &
          'Brown               ',  &
          'DarkRed             ',  &
          'Maroon              ',  &
          'SaddleBrown         ',  &
          'Sienna              ',  &
          'DarkGoldenrod       ',  &
          'Chocolate           ',  &
          'Peru                ',  &
          'Goldenrod           ',  &
          'Gold                ',  &
          'Yellow              ',  &
          'LightYellow         ',  &
          'LightGoldenrodYellow',  &       ! MS spelt it wrong!
          'LemonChiffon        ',  &
          'PapayaWhip          ',  &
          'Moccasin            ',  &
          'NavajoWhite         ',  &
          'Wheat               ',  &
          'BurlyWood           ',  &
          'Tan                 ',  &
          'PaleGoldenrod       ',  &
          'Khaki               ',  &
          'DarkKhaki           ' ] ) ]
    
    CALL create_hexagon(test%hexagon, hr)
    IF (hr /= 0) RETURN
    
  END SUBROUTINE construct_colour_test
  
  
  !*****************************************************************************
  !!
  !> Create the device dependent resources for the colour test.
  
  SUBROUTINE colour_CreateResources(test, render_target, hr)
    
    USE Direct2D
    USE Direct2DColor
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(colour_test), INTENT(INOUT) :: test
    !> The current render target.
    CLASS(ID2D1RenderTarget), INTENT(IN) :: render_target
    !> COM result code - negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !***************************************************************************
    
    CALL CreateSolidColorBrush(  &
        render_target,  &
        Black,  &
        test%brush,  &
        hr )
    IF (hr < 0) RETURN
    
  END SUBROUTINE colour_CreateResources
  
  
  !*****************************************************************************
  !!
  !> Discard the device dependent resources for the colour test.
  
  SUBROUTINE colour_DiscardResources(test)
    
    USE COMBase
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(colour_test), INTENT(INOUT) :: test
    
    !***************************************************************************
    
    IF (Exists(test%brush))           CALL Release(test%brush)
    
  END SUBROUTINE colour_DiscardResources
  
  
  !*****************************************************************************
  !!
  !> Oooh - pretty colours!
  !!
  !! Based on the image in the Direct2D API reference at 
  !! https://msdn.microsoft.com/en-us/library/windows/desktop/dd370907%28v=vs.85%29.aspx
  
  SUBROUTINE colour_Draw(test, render_target, hr)
    
    USE Direct2D
    USE Direct2DColor
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(colour_test), INTENT(INOUT) :: test
    !> The current render target.
    CLASS(ID2D1RenderTarget), INTENT(IN) :: render_target
    !> COM result code - negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: block          ! The index of the current block.
    LOGICAL :: right          ! If true, the item is drawn to the right.
    LOGICAL :: dark           ! If true, the item has a dark background.
    INTEGER :: item           ! The item index in the current block.
    
    !***************************************************************************
    
    ! Change from one unit being one dip (96 per inch) to one unit being 
    ! one millimetre.
    CALL SetTransform(  &
        render_target,  &
        Scale( &
          D2D1_SIZE_F(96.0_FLOAT / 25.4_FLOAT, 96.0_FLOAT / 25.4_FLOAT),  &
          D2D1_POINT_2F(0.0, 0.0) ) )
    
    CALL BeginDraw(render_target)
    
    CALL Clear(render_target, White)
    
    block_loop: DO block = 1, SIZE(test%colour_blocks)
      right = .FALSE.
      dark = .TRUE.
      
      DO item = 1, SIZE(test%colour_blocks(block)%names)
        CALL draw_swatch(  &
            render_target,  &
            block * 80.0_FLOAT - 40.0_FLOAT,  &
            item,  &
            test%colour_blocks(block)%names(item),  &
            right,  &
            dark,  &
            test,  &
            hr)
        IF (hr /= 0) EXIT block_loop
        right = .NOT. right
        IF (.NOT. right) dark = .NOT. dark
      END DO
    END DO block_loop
    
    CALL EndDraw(render_target, hr)
    
  END SUBROUTINE colour_Draw
  
  
  !*****************************************************************************
  !!
  !> Draw a colour swatch and associated colour information.
  
  SUBROUTINE draw_swatch( render_target, centreline_x, item, name, right,  &
      dark, test, hr )
    
    USE Direct2D
    USE Direct2DColor
    USE DirectWrite
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The render target to draw to.
    CLASS(ID2D1RenderTarget), INTENT(IN) :: render_target
    !> The centreline of the column of swatches.
    REAL(FLOAT), INTENT(IN) :: centreline_x
    !> The index of the swatch in the column.
    INTEGER, INTENT(IN) :: item
    !> The name of the colour.
    CHARACTER(*), INTENT(IN) :: name
    !> If true, the swatch is drawn to the right, otherwise left of the 
    !! centreline.
    LOGICAL, INTENT(IN) :: right
    !> If true, colour details are drawn on a dark background.
    LOGICAL, INTENT(IN) :: dark
    !> The test object.
    CLASS(colour_test), INTENT(IN) :: test
    !> COM result code, negative on error.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    !> The centre y ordinate of the swatch shape.
    REAL(FLOAT) :: centre_y
    !> The swatch shape transformed to its final position.
    TYPE(ID2D1TransformedGeometry) :: transformed_geometry
    !> The colour for the swatch.
    TYPE(D2D1_COLOR_F) :: the_colour
    !> The text format for the colour name and hex code.
    TYPE(IDWriteTextFormat) :: text_format
    !> The character representation of the hex value.
    CHARACTER(6) :: hex_value
    
    !***************************************************************************
    
    ! Create a copy of the geometry, translated to the relevant position.
    centre_y = item * 5.0_FLOAT
    CALL CreateTransformedGeometry(  &
        factory,  &
        test%hexagon,  &
        Translation( &
          centreline_x + MERGE(3.75, -3.75, right),  &
          centre_y ),  &
        transformed_geometry,  &
        hr )
    IF (hr /= 0) RETURN
    
    ! Draw the dark background if requested.
    IF (dark) THEN
      CALL SetColor(test%brush, Gainsboro)
      CALL FillRectangle(  &
          render_target,  &
          D2D1_RECT_F( &
            centreline_x + MERGE(5.0, -38.0, right),  &
            centre_y - 5.0_FLOAT,  &
            centreline_x + MERGE(38.0, -5.0, right),  &
            centre_y + 5.0_FLOAT ),  &
          test%brush )
    END IF
    
    ! Draw the line that separates name and hex code.
    CALL SetColor(test%brush, Silver)
    CALL DrawLine( render_target,  &
        D2D1_POINT_2F(centreline_x, centre_y),  &
        D2D1_POINT_2F(centreline_x + MERGE(38.0, -38.0, right), centre_y),  &
        test%brush,  &
        0.2_FLOAT )
    
    ! Lookup the colour, create a brush for that colour.
    the_colour = D2D1_COLOR_F_ifortbug(name)
    CALL SetColor(test%brush, the_colour)
    
    ! Draw our transformed swatch geometry.
    CALL FillGeometry(render_target, transformed_geometry, test%brush)
    CALL SetColor(test%brush, Silver)
    CALL DrawGeometry(  &
        render_target,  &
        transformed_geometry,  &
        test%brush,  &
        0.2_FLOAT )
    
    ! Create a text format for the name of the colour.
    CALL CreateTextFormat(  &
        write_factory,  &
        'Arial',  &
        DWRITE_FONT_WEIGHT_NORMAL,  &
        DWRITE_FONT_STYLE_NORMAL,  &
        DWRITE_FONT_STRETCH_NORMAL,  &
        3.0_FLOAT,  &
        'C',  &
        text_format,  &
        hr )
    IF (hr /= 0) RETURN
    CALL SetTextAlignment(  &
        text_format,  &
        MERGE( &
          DWRITE_TEXT_ALIGNMENT_LEADING,  &
          DWRITE_TEXT_ALIGNMENT_TRAILING,  &
          right ),  &
        hr )
    IF (hr /= 0) RETURN
    CALL SetParagraphAlignment(  &
        text_format,  &
        DWRITE_PARAGRAPH_ALIGNMENT_FAR,  &
        hr )
    IF (hr /= 0) RETURN
    CALL SetWordWrapping(  &
        text_format,  &
        DWRITE_WORD_WRAPPING_NO_WRAP,  &
        hr )
    IF (hr /= 0) RETURN
    
    ! Draw the colour name.
    CALL SetColor(test%brush, Black)
    CALL DrawText(  &
        render_target,  &
        TRIM(name),  &
        text_format,  &
        D2D1_RECT_F( &
          centreline_x + MERGE(10.0, -38.0, right),  &
          centre_y - 5.0_FLOAT,  &
          centreline_x + MERGE(38.0, -10.0, right),  &
          centre_y ),  &
        test%brush )
    
    ! Adjust the text format for the hex colour value.
    CALL SetParagraphAlignment(  &
        text_format,  &
        DWRITE_PARAGRAPH_ALIGNMENT_NEAR,  &
        hr )
    IF (hr /= 0) RETURN
    
    ! Determine and draw the hex value for the colour.
    WRITE (hex_value(1:2), "(Z2.2)") NINT(the_colour%r * 255.0)
    WRITE (hex_value(3:4), "(Z2.2)") NINT(the_colour%g * 255.0)
    WRITE (hex_value(5:6), "(Z2.2)") NINT(the_colour%b * 255.0)
    CALL DrawText(  &
        render_target,  &
        '#' // hex_value,  &
        text_format,  &
        D2D1_RECT_F( &
          centreline_x + MERGE(10.0, -38.0, right),  &
          centre_y,  &
          centreline_x + MERGE(38.0, -10.0, right),  &
          centre_y + 5.0_FLOAT ),  &
        test%brush )
    
  END SUBROUTINE draw_swatch
  
  
  !*****************************************************************************
  !!
  !> Create a filled hexagon geometry.
  
  SUBROUTINE create_hexagon(hexagon, hr)
    
    USE Direct2D
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The hexagon geometry object.
    TYPE(ID2D1PathGeometry), INTENT(OUT) :: hexagon
    !> HRESULT code, negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! The geometry sink used to assemble the path.
    TYPE(ID2D1GeometrySink) :: sink
    
    !***************************************************************************
    
    CALL CreatePathGeometry(factory, hexagon, hr)
    IF (hr < 0) RETURN
    CALL Open(hexagon, sink, hr)
    IF (hr < 0) RETURN
    CALL BeginFigure(  &
        sink,  &
        D2D1_POINT_2F(-2.5_FLOAT, -5.0_FLOAT),  &
        D2D1_FIGURE_BEGIN_FILLED )
    CALL AddLines(  &
        sink,  &
        [ D2D1_POINT_2F(2.5_FLOAT, -5.0_FLOAT),  &
          D2D1_POINT_2F(5.0_FLOAT, 0.0_FLOAT),  &
          D2D1_POINT_2F(2.5_FLOAT, 5.0_FLOAT),  &
          D2D1_POINT_2F(-2.5_FLOAT, 5.0_FLOAT),  &
          D2D1_POINT_2F(-5.0_FLOAT, 0.0_FLOAT) ] )
    CALL EndFigure(sink, D2D1_FIGURE_END_CLOSED)
    CALL Close(sink, hr)
    IF (hr < 0) RETURN
    
  END SUBROUTINE create_hexagon
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for transform_test
  
  
  !*****************************************************************************
  !!
  !> Construct a transform_test object.
  
  SUBROUTINE construct_transform_test(test, hr)
    
    USE Direct2D
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The constructed object.
    TYPE(transform_test), INTENT(OUT) :: test
    !> COM result code - negative on error.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !***************************************************************************
    
    CALL create_arrow(test%arrow, hr)
    IF (hr < 0) RETURN
    
  END SUBROUTINE construct_transform_test
  
  
  !*****************************************************************************
  !!
  !> Create the device dependent resources for the transform tests.
  
  SUBROUTINE transform_CreateResources(test, render_target, hr)
    
    USE Direct2D
    USE Direct2DColor
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(transform_test), INTENT(INOUT) :: test
    !> The current render target.
    CLASS(ID2D1RenderTarget), INTENT(IN) :: render_target
    !> COM result code - negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !***************************************************************************
    
    CALL CreateSolidColorBrush(  &
        render_target,  &
        Black,  &
        test%brush,  &
        hr )
    IF (hr < 0) RETURN
    
  END SUBROUTINE transform_CreateResources
  
  
  !*****************************************************************************
  !!
  !> Discard the device dependent resources for the transform tests.
  
  SUBROUTINE transform_DiscardResources(test)
    
    USE COMBase
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(transform_test), INTENT(INOUT) :: test
    
    !***************************************************************************
    
    IF (Exists(test%brush)) CALL Release(test%brush)
    
  END SUBROUTINE transform_DiscardResources
  
  
  !*****************************************************************************
  !!
  !> Some tests of the translation stuff.
  
  SUBROUTINE transform_Draw(test, render_target, hr)
    
    USE Direct2D
    USE Direct2DColor
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(transform_test), INTENT(INOUT) :: test
    !> The current render target.
    CLASS(ID2D1RenderTarget), INTENT(IN) :: render_target
    !> COM result code - negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !---------------------------------------------------------------------------
    ! Locals
    
    TYPE(ID2D1TransformedGeometry) :: transformed_arrow
    
    TYPE(D2D1_MATRIX_3X2_F) :: old_transform
    
    !***************************************************************************
    
    CALL BeginDraw(render_target)
    
    CALL SetTransform(render_target, IdentityMatrix)
    CALL Clear(render_target, White)
    
    drawing_operations: BLOCK
      
      !-------------------------------------------------------------------------
      ! Some really simple stuff, because we botched matrix multiply.
      
      CALL SetColor(test%brush, Black)
      CALL DrawRectangle(  &
          render_target,  &
          D2D1_RECT_F(-20.0_FLOAT, -20.0_FLOAT, 20.0_FLOAT, 20.0_FLOAT),  &
          test%brush )
      
      CALL SetTransform(render_target, Rotation(15.0_FLOAT))
      CALL SetColor(test%brush, Red)
      CALL DrawRectangle(  &
          render_target,  &
          D2D1_RECT_F(-20.0_FLOAT, -20.0_FLOAT, 20.0_FLOAT, 20.0_FLOAT),  &
          test%brush )
      
      CALL SetTransform(render_target, Translation(50.0_FLOAT, 0.0_FLOAT))
      CALL SetColor(test%brush, Blue)
      CALL DrawRectangle(  &
          render_target,  &
          D2D1_RECT_F(-20.0_FLOAT, -20.0_FLOAT, 20.0_FLOAT, 20.0_FLOAT),  &
          test%brush )
      
      CALL SetTransform(  &
          render_target,  &
          Translation(50.0_FLOAT, 0.0_FLOAT) * Rotation(15.0_FLOAT) )
      CALL SetColor(test%brush, Green)
      CALL DrawRectangle(  &
          render_target,  &
          D2D1_RECT_F(-20.0_FLOAT, -20.0_FLOAT, 20.0_FLOAT, 20.0_FLOAT),  &
          test%brush )
      
      CALL SetTransform(  &
          render_target,  &
          Rotation(15.0_FLOAT) * Translation(50.0_FLOAT, 0.0_FLOAT) )
      CALL SetColor(test%brush, Magenta)
      CALL DrawRectangle(  &
          render_target,  &
          D2D1_RECT_F(-20.0_FLOAT, -20.0_FLOAT, 20.0_FLOAT, 20.0_FLOAT),  &
          test%brush )
      
      CALL SetTransform(render_target, IdentityMatrix)
      
      !-------------------------------------------------------------------------
      ! Scale then shift
      
      CALL CreateTransformedGeometry(  &
          factory,  &
          test%arrow,  &
          Scale(2.0_FLOAT, 2.0_FLOAT, D2D1_POINT_2F())  &
            * Translation(50.0_FLOAT, 100.0_FLOAT),  &
          transformed_arrow,  &
          hr )
      IF (hr < 0) EXIT drawing_operations
      
      CALL SetColor(test%brush, Green)
      CALL FillGeometry(  &
          render_target,  &
          transformed_arrow,  &
          test%brush )
      CALL SetColor(test%brush, Black)
      CALL DrawGeometry(  &
          render_target,  &
          transformed_arrow,  &
          test%brush,  &
          1.0_FLOAT )
          
      old_transform = GetTransform(render_target)
      CALL SetTransform(  &
          render_target,  &
          Scale(2.0_FLOAT, 2.0_FLOAT, D2D1_POINT_2F())  &
            * Translation(50.0_FLOAT, 300.0_FLOAT) )
      
      CALL SetColor(test%brush, Blue)
      CALL FillGeometry(  &
          render_target,  &
          test%arrow,  &
          test%brush )
      CALL SetColor(test%brush, Black)
      CALL DrawGeometry(  &
          render_target,  &
          test%arrow,  &
          test%brush,  &
          1.0_FLOAT )
      
      CALL SetTransform(render_target, old_transform)
      
      !-------------------------------------------------------------------------
      ! Rotate, scale, then shift
      
      CALL CreateTransformedGeometry(  &
          factory,  &
          test%arrow,  &
          Rotation(45.0_FLOAT)  &
            * Scale(2.0_FLOAT, 2.0_FLOAT, D2D1_POINT_2F())  &
            * Translation(150.0_FLOAT, 100.0_FLOAT),  &
          transformed_arrow,  &
          hr )
      IF (hr < 0) EXIT drawing_operations
      
      CALL SetColor(test%brush, LightGreen)
      CALL FillGeometry(  &
          render_target,  &
          transformed_arrow,  &
          test%brush )
      CALL SetColor(test%brush, Black)
      CALL DrawGeometry(  &
          render_target,  &
          transformed_arrow,  &
          test%brush,  &
          1.0_FLOAT )
      
      old_transform = GetTransform(render_target)
      CALL SetTransform(  &
          render_target,  &
          Rotation(45.0_FLOAT)&
            * Scale(2.0_FLOAT, 2.0_FLOAT, D2D1_POINT_2F())  &
            * Translation(150.0_FLOAT, 300.0_FLOAT) )
      
      CALL SetColor(test%brush, LightBlue)
      CALL FillGeometry(  &
          render_target,  &
          test%arrow,  &
          test%brush )
      CALL SetColor(test%brush, Black)
      CALL DrawGeometry(  &
          render_target,  &
          test%arrow,  &
          test%brush,  &
          1.0_FLOAT )
      
      CALL SetTransform(render_target, old_transform)
      
      !-------------------------------------------------------------------------
      ! Rotate, big scale, then shift
      
      CALL CreateTransformedGeometry(  &
          factory,  &
          test%arrow,  &
          Rotation(-30.0_FLOAT)  &
            * Scale(5.0_FLOAT, 5.0_FLOAT, D2D1_POINT_2F())  &
            * Translation(250.0_FLOAT, 100.0_FLOAT),  &
          transformed_arrow,  &
          hr )
      IF (hr < 0) EXIT drawing_operations
      
      CALL SetColor(test%brush, DarkGreen)
      CALL FillGeometry(  &
          render_target,  &
          transformed_arrow,  &
          test%brush )
      CALL SetColor(test%brush, Black)
      CALL DrawGeometry(  &
          render_target,  &
          transformed_arrow,  &
          test%brush,  &
          1.0_FLOAT )
      
      old_transform = GetTransform(render_target)
      CALL SetTransform(  &
          render_target,  &
          Rotation(-30.0_FLOAT)  &
            * Scale(5.0_FLOAT, 5.0_FLOAT, D2D1_POINT_2F())  &
            * Translation(250.0_FLOAT, 300.0_FLOAT) )
      
      CALL SetColor(test%brush, DarkBlue)
      CALL FillGeometry(  &
          render_target,  &
          test%arrow,  &
          test%brush )
      CALL SetColor(test%brush, Black)
      CALL DrawGeometry(  &
          render_target,  &
          test%arrow,  &
          test%brush,  &
          1.0_FLOAT )
      
      CALL SetTransform(render_target, old_transform)
      
      !-------------------------------------------------------------------------
      ! Skew, scale, then shift
      
      CALL CreateTransformedGeometry(  &
          factory,  &
          test%arrow,  &
          Skew(-30.0_FLOAT, 0.0_FLOAT)  &
            * Scale(2.0_FLOAT, 2.0_FLOAT, D2D1_POINT_2F())  &
            * Translation(350.0_FLOAT, 100.0_FLOAT),  &
          transformed_arrow,  &
          hr )
      IF (hr < 0) EXIT drawing_operations
      
      CALL SetColor(test%brush, DarkGreen)
      CALL FillGeometry(  &
          render_target,  &
          transformed_arrow,  &
          test%brush )
      CALL SetColor(test%brush, Black)
      CALL DrawGeometry(  &
          render_target,  &
          transformed_arrow,  &
          test%brush,  &
          1.0_FLOAT )
      
      old_transform = GetTransform(render_target)
      CALL SetTransform(  &
          render_target,  &
          Skew(-30.0_FLOAT, 0.0_FLOAT)  &
            * Scale(2.0_FLOAT, 2.0_FLOAT, D2D1_POINT_2F())  &
            * Translation(350.0_FLOAT, 300.0_FLOAT) )
      
      CALL SetColor(test%brush, DarkBlue)
      CALL FillGeometry(  &
          render_target,  &
          test%arrow,  &
          test%brush )
      CALL SetColor(test%brush, Black)
      CALL DrawGeometry(  &
          render_target,  &
          test%arrow,  &
          test%brush,  &
          1.0_FLOAT )
      
      CALL SetTransform(render_target, old_transform)
      
    END BLOCK drawing_operations
    
    CALL EndDraw(render_target, hr)
    
  END SUBROUTINE transform_Draw
  
  
  !*****************************************************************************
  !!
  !> Create an arrow geometry.
  
  SUBROUTINE create_arrow(arrow, hr)
    
    USE Direct2D
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The arrow geometry object.
    CLASS(ID2D1Geometry), INTENT(OUT), ALLOCATABLE :: arrow
    !> HRESULT code, negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !---------------------------------------------------------------------------
    ! Locals
    
    TYPE(ID2D1PathGeometry) :: line
    TYPE(ID2D1EllipseGeometry) :: ellipse
    TYPE(ID2D1GeometryGroup), ALLOCATABLE :: group
    
    TYPE(ID2D1GeometrySink) :: sink
    
    !***************************************************************************
    
    CALL CreatePathGeometry(factory, line, hr)
    IF (hr < 0) RETURN
    
    CALL Open(line, sink, hr)
    IF (hr < 0) RETURN
    
    CALL BeginFigure(  &
        sink,  &
        D2D1_POINT_2F(0.0_FLOAT, -20.0_FLOAT),  &
        D2D1_FIGURE_BEGIN_HOLLOW )
    CALL AddLine(  &
        sink,  &
        D2D1_POINT_2F(0.0_FLOAT, 15.0_FLOAT) )
    CALL EndFigure(sink, D2D1_FIGURE_END_OPEN)
    
    CALL BeginFigure(  &
        sink,  &
        D2D1_POINT_2F(0.0_FLOAT, -20.0_FLOAT),  &
        D2D1_FIGURE_BEGIN_FILLED )
    CALL AddLines(  &
        sink,  &
        [ D2D1_POINT_2F(2.5_FLOAT, -25.0_FLOAT),  &
          D2D1_POINT_2F(-2.5_FLOAT, -25.0_FLOAT) ] )
    CALL EndFigure(sink, D2D1_FIGURE_END_CLOSED)
    
    CALL BeginFigure(  &
        sink,  &
        D2D1_POINT_2F(0.0_FLOAT, 25.0_FLOAT),  &
        D2D1_FIGURE_BEGIN_FILLED )
    CALL AddBezier(  &
        sink,  &
        D2D1_BEZIER_SEGMENT( &
          D2D1_POINT_2F(0.25_FLOAT, 23.0_FLOAT),  &
          D2D1_POINT_2F(1.5_FLOAT, 17.0_FLOAT),  &
          D2D1_POINT_2F(5.0_FLOAT, 15.0_FLOAT) ) )
    CALL AddLine(sink, D2D1_POINT_2f(-5.0_FLOAT, 15.0_FLOAT))
    CALL AddBezier(  &
        sink,  &
        D2D1_BEZIER_SEGMENT( &
          D2D1_POINT_2F(-1.5_FLOAT, 17.0_FLOAT),  &
          D2D1_POINT_2F(-0.25_FLOAT, 23.0_FLOAT),  &
          D2D1_POINT_2F(0.0_FLOAT, 25.0_FLOAT) ) )
    CALL EndFigure(sink, D2D1_FIGURE_END_CLOSED)
    CALL Close(sink, hr)
    IF (hr < 0) RETURN
    
    CALL CreateEllipseGeometry(  &
        factory,  &
        D2D1_ELLIPSE(  &
          D2D1_POINT_2F(0.0_FLOAT, 0.0_FLOAT),  &
          5.0_FLOAT,  &
          10.0_FLOAT ),  &
        ellipse,  &
        hr )
    IF (hr < 0) RETURN
    
    ALLOCATE(group)
    CALL CreateGeometryGroup(  &
        factory,  &
        D2D1_FILL_MODE_ALTERNATE,  &
        [ line%ID2D1Geometry, ellipse%ID2D1Geometry ],  &
        group,  &
        hr )
    IF (hr < 0) RETURN
    
    CALL MOVE_ALLOC(group, arrow)
    
  END SUBROUTINE create_arrow
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for clip_test
  
  
  !*****************************************************************************
  !!
  !> Construct a clip_test object.
  
  SUBROUTINE construct_clip_test(test, hr)
    
    USE Direct2D
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The constructed object.
    TYPE(clip_test), INTENT(OUT) :: test
    !> COM result code - negative on error.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !***************************************************************************
    
    CALL create_hexagon(test%hexagon, hr)
    IF (hr < 0) RETURN
    
    CALL CreateRectangleGeometry(  &
        factory,  &
        D2D1_RECT_F(210.0_FLOAT, 110.0_FLOAT, 290.0_FLOAT, 190.0_FLOAT),  &
        test%rectangle_geometry,  &
        hr )
    IF (hr < 0) RETURN
    
  END SUBROUTINE construct_clip_test
  
  
  !*****************************************************************************
  !!
  !> Create the device dependent resources for the clip tests.
  
  SUBROUTINE clip_CreateResources(test, render_target, hr)
    
    USE Direct2D
    USE Direct2DColor
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(clip_test), INTENT(INOUT) :: test
    !> The current render target.
    CLASS(ID2D1RenderTarget), INTENT(IN) :: render_target
    !> COM result code - negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !***************************************************************************
    
    CALL CreateSolidColorBrush(  &
        render_target,  &
        Black,  &
        test%brush,  &
        hr )
    IF (hr < 0) RETURN
    
    CALL CreateLayer(  &
        render_target,  &
        test%layer,  &
        hr,  &
        D2D1_SIZE_F(100.0_FLOAT, 100.0_FLOAT) )
    IF (hr < 0) RETURN
    
  END SUBROUTINE clip_CreateResources
  
  
  !*****************************************************************************
  !!
  !> Discard the device dependent resources for the clip tests.
  
  SUBROUTINE clip_DiscardResources(test)
    
    USE COMBase
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(clip_test), INTENT(INOUT) :: test
    
    !***************************************************************************
    
    IF (Exists(test%brush)) CALL Release(test%brush)
    IF (Exists(test%layer)) CALL Release(test%layer)
    
  END SUBROUTINE clip_DiscardResources
  
  
  !*****************************************************************************
  !!
  !> Some tests of the clipping stuff.
  
  SUBROUTINE clip_Draw(test, render_target, hr)
    
    USE Direct2D
    USE Direct2DColor
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(clip_test), INTENT(INOUT) :: test
    !> The current render target.
    CLASS(ID2D1RenderTarget), INTENT(IN) :: render_target
    !> COM result code - negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(D2D1_MATRIX_3X2_F) :: old_transform
    
    !***************************************************************************
    
    CALL SetTransform(render_target, IdentityMatrix)
    
    CALL BeginDraw(render_target)
    
    drawing_operations: BLOCK
      
      CALL Clear(render_target, White)
      
      !---------------------------------------------------------------------------
      ! Axis aligned clip with axis aligned transform
      
      CALL SetColor(test%brush, Red)
      CALL DrawEllipse(  &
          render_target,  &
          D2D1_ELLIPSE( &
            D2D1_POINT_2F(50.0_FLOAT, 50.0_FLOAT),  &
            60.0_FLOAT,  &
            40.0_FLOAT ),  &
          test%brush,  &
          2.0_FLOAT )
      
      CALL PushAxisAlignedClip(  &
          render_target,  &
          D2D1_RECT_F(10.0_FLOAT, 10.0_FLOAT, 90.0_FLOAT, 90.0_FLOAT),  &
          D2D1_ANTIALIAS_MODE_PER_PRIMITIVE )
      
      CALL SetColor(test%brush, Silver)
      CALL FillEllipse(  &
          render_target,  &
          D2D1_ELLIPSE( &
            D2D1_POINT_2F(50.0_FLOAT, 50.0_FLOAT),  &
            60.0_FLOAT,  &
            40.0_FLOAT ),  &
          test%brush )
      
      CALL PopAxisAlignedClip(render_target)
      
      !---------------------------------------------------------------------------
      ! Axis aligned clip with non-axis aligned transform
      
      old_transform = GetTransform(render_target)
      CALL SetTransform(  &
          render_target,  &
          Rotation(10.0_FLOAT) * Translation(0.0_FLOAT, -40.0_FLOAT) )
      
      CALL SetColor(test%brush, Green)
      CALL DrawEllipse(  &
          render_target,  &
          D2D1_ELLIPSE( &
            D2D1_POINT_2F(250.0_FLOAT, 50.0_FLOAT),  &
            60.0_FLOAT,  &
            40.0_FLOAT ),  &
          test%brush,  &
          2.0_FLOAT )
      
      CALL SetColor(test%brush, LightGreen)
      CALL DrawRectangle(  &
          render_target,  &
          D2D1_RECT_F(210.0_FLOAT, 10.0_FLOAT, 290.0_FLOAT, 90.0_FLOAT),  &
          test%brush )
      
      CALL PushAxisAlignedClip(  &
          render_target,  &
          D2D1_RECT_F(210.0_FLOAT, 10.0_FLOAT, 290.0_FLOAT, 90.0_FLOAT),  &
          D2D1_ANTIALIAS_MODE_PER_PRIMITIVE )
      
      CALL SetColor(test%brush, Silver)
      CALL FillEllipse(  &
          render_target,  &
          D2D1_ELLIPSE( &
            D2D1_POINT_2F(250.0_FLOAT, 50.0_FLOAT),  &
            60.0_FLOAT,  &
            40.0_FLOAT ),  &
          test%brush )
      
      CALL PopAxisAlignedClip(render_target)
      CALL SetTransform(render_target, old_transform)
      
      !---------------------------------------------------------------------------
      ! Clipping using a layer
      
      CALL SetColor(test%brush, Blue)
      CALL DrawEllipse(  &
          render_target,  &
          D2D1_ELLIPSE( &
            D2D1_POINT_2F(50.0_FLOAT, 150.0_FLOAT),  &
            60.0_FLOAT,  &
            40.0_FLOAT ),  &
          test%brush,  &
          2.0_FLOAT )
      ! Rectangle in the middle to demonstrate opacity
      CALL FillRectangle(  &
          render_target,  &
          D2D1_RECT_F(40.0_FLOAT, 140.0_FLOAT, 60.0_FLOAT, 160.0_FLOAT),  &
          test%brush )
      
      CALL PushLayer(  &
          render_target,  &
          D2D1_LAYER_PARAMETERS( &
            GEOMETRICMASK=test%hexagon,  &
            MASKTRANSFORM=Scale(10.0_FLOAT, 10.0_FLOAT, D2D1_POINT_2F())  &
              * Translation(50.0_FLOAT, 150.0_FLOAT),  &
            OPACITY=0.5_FLOAT ),  &
          test%layer )
      
      CALL SetColor(test%brush, Coral)
      CALL FillEllipse(  &
          render_target,  &
          D2D1_ELLIPSE( &
            D2D1_POINT_2F(50.0_FLOAT, 150.0_FLOAT),  &
            60.0_FLOAT,  &
            40.0_FLOAT ),  &
          test%brush )
      
      CALL PopLayer(render_target)
      
      !---------------------------------------------------------------------------
      ! Clipping using a layer - with rotation
      
      old_transform = GetTransform(render_target)
      CALL SetTransform(  &
          render_target,  &
          Rotation(10.0_FLOAT) * Translation(0.0_FLOAT, -40.0_FLOAT) )
      
      CALL SetColor(test%brush, Green)
      CALL DrawEllipse(  &
          render_target,  &
          D2D1_ELLIPSE( &
            D2D1_POINT_2F(250.0_FLOAT, 150.0_FLOAT),  &
            60.0_FLOAT,  &
            40.0_FLOAT ),  &
          test%brush,  &
          2.0_FLOAT )
      CALL FillRectangle(  &
          render_target,  &
          D2D1_RECT_F(240.0_FLOAT, 140.0_FLOAT, 260.0_FLOAT, 160.0_FLOAT),  &
          test%brush )
      
      CALL SetColor(test%brush, LightGreen)
      CALL DrawRectangle(  &
          render_target,  &
          D2D1_RECT_F(210.0_FLOAT, 110.0_FLOAT, 290.0_FLOAT, 190.0_FLOAT),  &
          test%brush )
      
      CALL PushLayer(  &
          render_target,  &
          D2D1_LAYER_PARAMETERS( &
            GEOMETRICMASK=test%rectangle_geometry,  &
            OPACITY=0.5_FLOAT ),  &
          test%layer )
      
      CALL SetColor(test%brush, Coral)
      CALL FillEllipse(  &
          render_target,  &
          D2D1_ELLIPSE( &
            D2D1_POINT_2F(250.0_FLOAT, 150.0_FLOAT),  &
            60.0_FLOAT,  &
            40.0_FLOAT ),  &
          test%brush )
      
      CALL PopLayer(render_target)
      
      CALL SetTransform(render_target, old_transform)
    
    !---------------------------------------------------------------------------
    
    END BLOCK drawing_operations
      
    CALL EndDraw(render_target, hr)
    
  END SUBROUTINE clip_Draw
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for line_test.
  
  
  !*****************************************************************************
  !!
  !> Construct a line_test object.
  
  SUBROUTINE construct_line_test(test, hr)
    
    USE Direct2D
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The constructed object.
    TYPE(line_test), INTENT(OUT) :: test
    !> COM result code - negative on error.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !***************************************************************************
    
    ALLOCATE(test%lines(1000))
    
    CALL RANDOM_NUMBER(test%lines%start%x)
    CALL RANDOM_NUMBER(test%lines%start%y)
    
    CALL RANDOM_NUMBER(test%lines%finish%x)
    CALL RANDOM_NUMBER(test%lines%finish%y)
    
    CALL RANDOM_NUMBER(test%lines%colour%r)
    CALL RANDOM_NUMBER(test%lines%colour%b)
    CALL RANDOM_NUMBER(test%lines%colour%g)
    CALL RANDOM_NUMBER(test%lines%colour%a)
    test%lines%colour%a = test%lines%colour%a * 0.5_FLOAT + 0.5_FLOAT
    
    CALL RANDOM_NUMBER(test%lines%width)
    test%lines%width = test%lines%width * 0.01_FLOAT
    
    hr = 0_HRESULT
    
  END SUBROUTINE construct_line_test
  
  
  !*****************************************************************************
  !!
  !> Create the device dependent resources for the line test.
  
  SUBROUTINE line_CreateResources(test, render_target, hr)
    
    USE Direct2D
    USE Direct2DColor
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(line_test), INTENT(INOUT) :: test
    !> The current render target.
    CLASS(ID2D1RenderTarget), INTENT(IN) :: render_target
    !> COM result code - negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !***************************************************************************
    
    CALL CreateSolidColorBrush(  &
        render_target,  &
        Black,  &
        test%brush,  &
        hr )
    IF (hr < 0) RETURN
    
  END SUBROUTINE line_CreateResources
  
  
  !*****************************************************************************
  !!
  !> Discard the device dependent resources for the line test.
  
  SUBROUTINE line_DiscardResources(test)
    
    USE COMBase
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(line_test), INTENT(INOUT) :: test
    
    !***************************************************************************
    
    IF (Exists(test%brush))           CALL Release(test%brush)
    
  END SUBROUTINE line_DiscardResources
  
  
  !*****************************************************************************
  !!
  !> A test that draws a lot of lines.
  
  SUBROUTINE line_Draw(test, render_target, hr)
    
    USE Direct2D
    USE DirectWrite
    USE Direct2DColor
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(line_test), INTENT(INOUT) :: test
    !> The current render target.
    CLASS(ID2D1RenderTarget), INTENT(IN) :: render_target
    !> COM result code - negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: iline          ! line index.
    
    !***************************************************************************
    
    CALL SetTransform(  &
        render_target,  &
        Scale(GetSize(render_target), D2D1_POINT_2F()) )
    
    CALL BeginDraw(render_target)
    CALL Clear(render_target, White)
    
    DO iline = 1, SIZE(test%lines)
      CALL SetColor(test%brush, test%lines(iline)%colour)
      CALL DrawLine(  &
          render_target,  &
          test%lines(iline)%start,  &
          test%lines(iline)%finish,  &
          test%brush,  &
          test%lines(iline)%width )
    END DO
    
    CALL EndDraw(render_target, hr)
    
  END SUBROUTINE line_Draw
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for ellipse_test.
  
  
  !*****************************************************************************
  !!
  !> Construct a ellipse_test object.
  
  SUBROUTINE construct_ellipse_test(test, hr)
    
    USE Direct2D
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    !> The constructed object.
    TYPE(ellipse_test), INTENT(OUT) :: test
    !> COM result code - negative on error.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !***************************************************************************
    
    ALLOCATE(test%ellipses(1000))
    
    CALL RANDOM_NUMBER(test%ellipses%ellipse%point%x)
    CALL RANDOM_NUMBER(test%ellipses%ellipse%point%y)
    
    CALL RANDOM_NUMBER(test%ellipses%ellipse%radiusX)
    test%ellipses%ellipse%radiusX = test%ellipses%ellipse%radiusX * 0.1_FLOAT
    CALL RANDOM_NUMBER(test%ellipses%ellipse%radiusY)
    test%ellipses%ellipse%radiusY = test%ellipses%ellipse%radiusY * 0.1_FLOAT
    
    CALL RANDOM_NUMBER(test%ellipses%border_colour%r)
    CALL RANDOM_NUMBER(test%ellipses%border_colour%b)
    CALL RANDOM_NUMBER(test%ellipses%border_colour%g)
    CALL RANDOM_NUMBER(test%ellipses%border_colour%a)
    test%ellipses%border_colour%a  &
        = test%ellipses%border_colour%a * 0.5_FLOAT + 0.5_FLOAT
    
    CALL RANDOM_NUMBER(test%ellipses%width)
    test%ellipses%width = test%ellipses%width * 0.01_FLOAT
    
    CALL RANDOM_NUMBER(test%ellipses%fill_colour%r)
    CALL RANDOM_NUMBER(test%ellipses%fill_colour%b)
    CALL RANDOM_NUMBER(test%ellipses%fill_colour%g)
    CALL RANDOM_NUMBER(test%ellipses%fill_colour%a)
    test%ellipses%fill_colour%a  &
        = test%ellipses%fill_colour%a * 0.5_FLOAT + 0.5_FLOAT
    
    hr = 0_HRESULT
    
  END SUBROUTINE construct_ellipse_test
  
  
  !*****************************************************************************
  !!
  !> Create the device dependent resources for the ellipse test.
  
  SUBROUTINE ellipse_CreateResources(test, render_target, hr)
    
    USE Direct2D
    USE Direct2DColor
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(ellipse_test), INTENT(INOUT) :: test
    !> The current render target.
    CLASS(ID2D1RenderTarget), INTENT(IN) :: render_target
    !> COM result code - negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !***************************************************************************
    
    CALL CreateSolidColorBrush(  &
        render_target,  &
        Black,  &
        test%brush,  &
        hr )
    IF (hr < 0) RETURN
    
  END SUBROUTINE ellipse_CreateResources
  
  
  !*****************************************************************************
  !!
  !> Discard the device dependent resources for the ellipse test.
  
  SUBROUTINE ellipse_DiscardResources(test)
    
    USE COMBase
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(ellipse_test), INTENT(INOUT) :: test
    
    !***************************************************************************
    
    IF (Exists(test%brush))           CALL Release(test%brush)
    
  END SUBROUTINE ellipse_DiscardResources
  
  
  !*****************************************************************************
  !!
  !> A test that draws lots of ellipses.
  
  SUBROUTINE ellipse_Draw(test, render_target, hr)
    
    USE Direct2D
    USE DirectWrite
    USE Direct2DColor
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(ellipse_test), INTENT(INOUT) :: test
    !> The current render target.
    CLASS(ID2D1RenderTarget), INTENT(IN) :: render_target
    !> COM result code - negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: iel            ! ellipse index.
    
    !***************************************************************************
    
    CALL SetTransform(  &
        render_target,  &
        Scale(GetSize(render_target), D2D1_POINT_2F()) )
    
    CALL BeginDraw(render_target)
    CALL Clear(render_target, White)
    
    DO iel = 1, SIZE(test%ellipses)
      CALL SetColor(test%brush, test%ellipses(iel)%fill_colour)
      CALL FillEllipse(  &
          render_target,  &
          test%ellipses(iel)%ellipse,  &
          test%brush )
      
      CALL SetColor(test%brush, test%ellipses(iel)%border_colour)
      CALL DrawEllipse(  &
          render_target,  &
          test%ellipses(iel)%ellipse,  &
          test%brush,  &
          test%ellipses(iel)%width )
    END DO
    
    CALL EndDraw(render_target, hr)
    
  END SUBROUTINE ellipse_Draw
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bmpbrush_test
  
  
  !*****************************************************************************
  !!
  !> Construct a bmbrush_test object.
  
  SUBROUTINE construct_bmpbrush_test(test, hr)
    
    USE Direct2D
    USE WinImagingComponent, WIC_GetSize => GetSize
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The constructed object.
    TYPE(bmpbrush_test), INTENT(OUT) :: test
    !> COM result code - negative on error.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !---------------------------------------------------------------------------
    ! Locals
    
    TYPE(ID2D1GeometrySink) :: sink
    TYPE(IWICBitmapDecoder) :: bitmap_decoder
    TYPE(IWICBitmapFrameDecode) :: bitmap_frame
    
    !***************************************************************************
    
    hr = 0_HRESULT
    
    CALL CreatePathGeometry(  &
        factory,  &
        test%fill_shape,  &
        hr )
    IF (hr < 0) RETURN
    
    CALL Open(test%fill_shape, sink, hr)
    IF (hr < 0) RETURN
    
    CALL BeginFigure(  &
        sink,  &
        D2D1_POINT_2F(100.0_FLOAT, 250.0_FLOAT),  &
        D2D1_FIGURE_BEGIN_FILLED )
    CALL AddLines(  &
        sink,  &
        [ D2D1_POINT_2F(175.0_FLOAT, 350.0_FLOAT),  &
          D2D1_POINT_2F(25.0_FLOAT, 350.0_FLOAT) ] )
    CALL EndFigure(sink, D2D1_FIGURE_END_CLOSED)
    
    CALL Close(sink, hr)
    IF (hr < 0) RETURN
    
    !---------------------------------------------------------------------------
    ! Load the test bitmap file using Windows Imaging Component.
    
    ! Create the Windows Imaging Component factory.
    CALL WICCreateFactory(test%wic_factory, hr)
    IF (hr < 0) RETURN
    
    ! Load the test bitmap.
    CALL CreateDecoderFromFilename(  &
        test%wic_factory,  &
        'chook.jpg',  &
        GENERIC_READ,  &
        WICDecodeMetadataCacheOnDemand,  &
        bitmap_decoder,  &
        hr )
    IF (hr < 0) RETURN
    
    ! Get the relevant frame (the test file only has one frame).
    CALL GetFrame(  &
        bitmap_decoder,  &
        0,  &
        bitmap_frame,  &
        hr )
    IF (hr < 0) RETURN
    
    ! Create a converter source to allow accessing the bitmap in a 
    ! Direct2D compatible format.  This is the source for the 
    ! actual Direct2D bitmap.
    BLOCK
      TYPE(IWICFormatConverter), ALLOCATABLE :: converter_source
      
      ALLOCATE(converter_source)
      CALL CreateFormatConverter(test%wic_factory, converter_source, hr)
      IF (hr < 0) RETURN
    
      CALL Initialize(  &
          converter_source,  &
          bitmap_frame,  &
          GUID_WICPixelFormat32bppPBGRA,  &
          hr );
      IF (hr < 0) RETURN
      
      CALL MOVE_ALLOC(converter_source, test%wic_bitmap_source)
    END BLOCK
    
  END SUBROUTINE construct_bmpbrush_test
  
  
  !*****************************************************************************
  !!
  !> Create the device dependent resources for the bmpbrush test.
  
  SUBROUTINE bmpbrush_CreateResources(test, render_target, hr)
    
    USE Direct2D
    USE Direct2DColor
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(bmpbrush_test), INTENT(INOUT) :: test
    !> The current render target.
    CLASS(ID2D1RenderTarget), INTENT(IN) :: render_target
    !> COM result code - negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! A compatible render target to the render target for the main display.
    TYPE(ID2D1BitmapRenderTarget) :: bitmap_target
    
    ! A brush on the compatible target.  Because the targets are compatible, 
    ! we could also have created this brush on the main render target.
    TYPE(ID2D1SolidColorBrush) :: bitmap_target_brush
    
    ! The bitmap that the bitmap render target drew to.
    TYPE(ID2D1Bitmap) :: bitmap
    
    ! Transform (ifort bug workaround).
    TYPE(D2D1_MATRIX_3X2_F) :: txfrm
    
    !***************************************************************************
    
    !---------------------------------------------------------------------------
    ! Create the brush for the drawn bitmap.
    
    CALL CreateCompatibleRenderTarget(  &
        render_target,  &
        bitmap_target,  &
        hr,  &
        DESIREDSIZE=D2D1_SIZE_F(100.0_FLOAT, 100.0_FLOAT) )
    IF (hr < 0) RETURN
    
    CALL CreateSolidColorBrush(  &
        bitmap_target,  &
        Lime,  &
        bitmap_target_brush,  &
        hr )
    IF (hr < 0) RETURN
    
    ! Draw the patten into the bitmap.
    CALL BeginDraw(bitmap_target)
    drawing_operations: BLOCK
      CALL Clear(bitmap_target, DarkOliveGreen)
      
      CALL DrawLine(  &
          bitmap_target,  &
          D2D1_POINT_2F(33.3_FLOAT, 0.0_FLOAT),  &
          D2D1_POINT_2F(33.3_FLOAT, 100.0_FLOAT),  &
          bitmap_target_brush )
      
      CALL DrawLine(  &
          bitmap_target,  &
          D2D1_POINT_2F(66.7_FLOAT, 0.0_FLOAT),  &
          D2D1_POINT_2F(66.7_FLOAT, 100.0_FLOAT),  &
          bitmap_target_brush )
      
      CALL DrawLine(  &
          bitmap_target,  &
          D2D1_POINT_2F(0.0_FLOAT, 33.3_FLOAT),  &
          D2D1_POINT_2F(100.0_FLOAT, 33.3_FLOAT),  &
          bitmap_target_brush )
      
      CALL DrawLine(  &
          bitmap_target,  &
          D2D1_POINT_2F(0.0_FLOAT, 66.7_FLOAT),  &
          D2D1_POINT_2F(100.0_FLOAT, 66.7_FLOAT),  &
          bitmap_target_brush )
    END BLOCK drawing_operations
    
    CALL EndDraw(bitmap_target, hr)
    
    CALL GetBitmap(bitmap_target, bitmap, hr)
    IF (hr < 0) RETURN
    
    ! Create a bitmap brush based on that bitmap.
    
    ! ifort thinks Scale in the following is intrinsic if the expression 
    ! is used in the structure constructor.
    txfrm = Scale(0.5_FLOAT, 0.5_FLOAT, D2D1_POINT_2F())  &
        * Rotation(25.0_FLOAT)
    CALL CreateBitmapBrush(  &
        render_target,  &
        bitmap,  &
        test%bitmap_brush,  &
        hr,  &
        bitmapBrushProperties=D2D1_BITMAP_BRUSH_PROPERTIES(  &
          extendModeX=D2D1_EXTEND_MODE_WRAP,  &
          extendModeY=D2D1_EXTEND_MODE_WRAP ),  &
        brushProperties=D2D1_BRUSH_PROPERTIES(TRANSFORM=txfrm) )
    IF (hr < 0) RETURN
    
    ! A solid brush for outlines.
    CALL CreateSolidColorBrush(  &
        render_target,  &
        Black,  &
        test%solid_brush,  &
        hr )
    IF (hr < 0) RETURN
    
    !---------------------------------------------------------------------------
    ! Create the brush for the bitmap from the file.
    
    ! Create a D2D1 bitmap variant of the file bitmap.
    CALL CreateBitmapFromWICBitmap(  &
        render_target,  &
        test%wic_bitmap_source,  &
        test%chook,  &
        hr )
    IF (hr < 0) RETURN
    
    ! Create a bitmap brush from the file bitmap.
    CALL CreateBitmapBrush(  &
        render_target,  &
        test%chook,  &
        test%chook_brush,  &
        hr,  &
        bitmapBrushProperties=D2D1_BITMAP_BRUSH_PROPERTIES(  &
          extendModeX=D2D1_EXTEND_MODE_WRAP,  &
          extendModeY=D2D1_EXTEND_MODE_WRAP ) )
    IF (hr < 0) RETURN
    
  END SUBROUTINE bmpbrush_CreateResources
  
  
  !*****************************************************************************
  !!
  !> Discard the device dependent resources for the bmpbrush test.
  
  SUBROUTINE bmpbrush_DiscardResources(test)
    
    USE COMBase
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(bmpbrush_test), INTENT(INOUT) :: test
    
    !***************************************************************************
    
    IF (Exists(test%bitmap_brush))  &
        CALL Release(test%bitmap_brush)
    
    IF (Exists(test%solid_brush))  &
        CALL Release(test%solid_brush)
    
    IF (Exists(test%chook))  &
        CALL Release(test%chook)
    
    IF (Exists(test%chook_brush))  &
        CALL Release(test%chook_brush)
    
  END SUBROUTINE bmpbrush_DiscardResources
  
  
  !*****************************************************************************
  !!
  !> A test of bitmap brushes and related APIs.
  
  SUBROUTINE bmpbrush_Draw(test, render_target, hr)
    
    USE Direct2D
    USE DirectWrite
    USE Direct2DColor
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(bmpbrush_test), INTENT(INOUT) :: test
    !> The current render target.
    CLASS(ID2D1RenderTarget), INTENT(IN) :: render_target
    !> COM result code - negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !***************************************************************************
    
    hr = 0_HRESULT
    
    CALL SetTransform(render_target, IdentityMatrix)
    
    CALL BeginDraw(render_target)
    drawing_operations: BLOCK
      CALL Clear(render_target, White)
      CALL FillGeometry(  &
          render_target,  &
          test%fill_shape,  &
          test%bitmap_brush )
      CALL DrawGeometry(  &
          render_target,  &
          test%fill_shape,  &
          test%solid_brush )
      
      ! Draw the bitmap with defaults.
      CALL DrawBitmap(  &
          render_target,  &
          test%chook )
      
      ! Draw to a certain rectangle
      CALL DrawBitmap(  &
          render_target,  &
          test%chook,  &
          DESTINATIONRECTANGLE=D2D1_RECT_F_IfortBug(  &
            D2D1_POINT_2F(250.0_FLOAT, 20.0_FLOAT),  &
            D2D1_SIZE_F(50.0_FLOAT, 50.0_FLOAT) ) )
      
      ! Draw from a certain rectangle.
      CALL DrawBitmap(  &
          render_target,  &
          test%chook,  &
          SOURCERECTANGLE=D2D1_RECT_F_IfortBug(  &
            D2D1_POINT_2F(180.0_FLOAT, 35.0_FLOAT),  &
            D2D1_SIZE_F(50.0_FLOAT, 50.0_FLOAT) ) )
      
      ! Use as a brush.
      BLOCK
        TYPE(IDWriteTextFormat) :: text_format
        
        CALL CreateTextFormat(  &
            write_factory,  &
            'Arial',  &
            DWRITE_FONT_WEIGHT_BOLD,  &
            DWRITE_FONT_STYLE_NORMAL,  &
            DWRITE_FONT_STRETCH_NORMAL,  &
            200.0_FLOAT,  &
            'C',  &
            text_format,  &
            hr )
        IF (hr < 0) RETURN
        
        CALL DrawText(  &
            render_target,  &
            'SUPER' // ACHAR(10) // 'CHOOK',  &
            text_format,  &
            D2D1_RECT_F_IfortBug(  &
              D2D1_POINT_2F(400.0_FLOAT, 10.0_FLOAT),  &
              D2D1_SIZE_F(800.0_FLOAT, 200.0_FLOAT) ),  &
            test%chook_brush )
      END BLOCK
      
    END BLOCK drawing_operations
    
    CALL EndDraw(render_target, hr)
    
  END SUBROUTINE bmpbrush_Draw
  
  
  ! Same story as for logo_final_xxx
  subroutine bmpbrush_final_suspect_ifort_bug(obj)
    use combase
    type(bmpbrush_test), intent(inout) :: obj
    
    call release(obj%fill_shape)
  end subroutine bmpbrush_final_suspect_ifort_bug
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for logo test.
  
  
  !*****************************************************************************
  !!
  !> Construct a logo_test object.
  
  SUBROUTINE construct_logo_test(test, hr)
    
    USE Direct2D
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The constructed object.
    TYPE(logo_test), INTENT(OUT) :: test
    !> COM result code - negative on error.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Geometry sink for constructing paths.
    TYPE(ID2D1GeometrySink) :: sink
    
    ! Current point for construction based on relative movements.
    TYPE(D2D1_POINT_2F) :: point
    
    ! Untransformed ellipse geometry
    TYPE(ID2D1EllipseGeometry) :: ellipse
    
    ! The bezier curves that make up the M.
    REAL, PARAMETER :: m1_curves(2,3,44) = RESHAPE(  &
        [ -22,32      , -45,43      , -67,32    ,  &
          -167,-96    , -324,-193   , -480,-289 ,  &
          -34,-21     , -34,-53     , -11,-86   ,  &
          67,-128     , 167,-310    , 301,-556  ,  &
          134,-246    , 235,-428    , 302,-546  ,  &
          11,-43      , 22,-64      , 11,-64    ,  &
          -11,0       , -168,160    , -480,460  ,  &
          -179,182    , -346,353    , -525,525  ,  &
          -22,21      , -45,32      , -56,21    ,  &
          -11,0       , -22,-11     , -56,-32   ,  &
          -44,-32     , -67,-54     , -67,-86   ,  &
          23,-160     , 56,-385     , 101,-663  ,  &
          44,-300     , 66,-514     , 78,-653   ,  &
          0,0         , 0,-11       , 0,-22     ,  &
          -12,11      , -22,22      , -34,43    ,  &
          -56,107     , -134,279    , -245,493  ,  &
          -112,224    , -190,385    , -257,492  ,  &
          -12,43      , -34,53      , -67,43    ,  &
          -146,-97    , -291,-204   , -436,-300 ,  &
          -22,-11     , -22,-22     , -22,-32   ,  &
          0,-11       , 0,-21       , 11,-32    ,  &
          56,-118     , 145,-289    , 279,-503  ,  &
          123,-225    , 223,-396    , 279,-503  ,  &
          67,-129     , 179,-353    , 336,-686  ,  &
          44,-85      , 111,-224    , 189,-396  ,  &
          11,-10      , 11,-10      , 11,-21    ,  &
          12,-21      , 45,-32      , 78,-10    ,  &
          68,10       , 146,42      , 258,64    ,  &
          111,32      , 200,64      , 257,85    ,  &
          11,0        , 11,33       , 11,75     ,  &
          -45,418     , -90,835     , -146,1274 ,  &
          -11,85      , 0,139       , 23,150    ,  &
          0,0         , 55,-32      , 133,-107  ,  &
          224,-203    , 570,-514    , 1028,-953 ,  &
          45,-42      , 79,-64      , 101,-53   ,  &
          22,11       , 67,21       , 122,53    ,  &
          57,21       , 101,43      , 135,54    ,  &
          22,10       , 66,21       , 123,32    ,  &
          44,11       , 89,21       , 111,32    ,  &
          34,11       , 45,32       , 23,64     ,  &
          -79,150     , -201,375    , -358,664  ,  &
          -168,300    , -290,524    , -369,674  ,  &
          -78,139     , -178,343    , -312,632  ,  &
          -134,278    , -246,492    , -313,631 ],  &
        [2, 3, 44 ] ) / 100.0_FLOAT
    
    REAL, PARAMETER :: e_curves(2,3,48) = RESHAPE(  &
        [ -33,54      , -100,161    , -201,322  ,  &
          -111,160    , -179,256    , -212,288  ,  &
          -22,22      , -45,22      , -67,11    ,  &
          -11,0       , -23,-11     , -34,-21   ,  &
          -67,-54     , -256,-161   , -591,-343 ,  &
          -68,-43     , -179,-96    , -313,-182 ,  &
          -145,-75    , -246,-139   , -313,-182 ,  &
          -33,-10     , -44,-42     , -22,-85   ,  &
          78,-150     , 201,-364    , 368,-664  ,  &
          168,-289    , 291,-514    , 369,-663  ,  &
          78,-150     , 223,-439    , 435,-857  ,  &
          45,-107     , 123,-267    , 224,-481  ,  &
          11,-11      , 11,-22      , 22,-32    ,  &
          11,-33      , 33,-43      , 56,-33    ,  &
          78,33       , 190,75      , 324,139   ,  &
          145,65      , 257,118     , 324,140   ,  &
          78,32       , 179,74      , 335,128   ,  &
          145,54      , 257,107     , 324,128   ,  &
          56,33       , 67,65       , 44,129    ,  &
          -11,32      , -55,128     , -134,300  ,  &
          -67,171     , -123,267    , -133,310  ,  &
          -23,22      , -34,32      , -57,32    ,  &
          0,-10       , -11,-10     , -33,-21   ,  &
          -235,-150   , -458,-268   , -670,-354 ,  &
          -45,-31     , -90,-21     , -123,0    ,  &
          -11,22      , -33,76      , -89,161   ,  &
          -45,96      , -67,161     , -67,193   ,  &
          11,42       , 33,64       , 78,85     ,  &
          11,11       , 33,11       , 56,22     ,  &
          33,11       , 55,21       , 78,32     ,  &
          112,54      , 246,107     , 369,160   ,  &
          11,0        , 44,11       , 89,22     ,  &
          33,11       , 55,21       , 78,32     ,  &
          22,11       , 22,32       , 11,64     ,  &
          11,-32      , -45,75      , -190,311  ,  &
          -67,106     , -134,214    , -190,331  ,  &
          -11,22      , -33,33      , -67,22    ,  &
          0,0         , -33,-32     , -111,-86  ,  &
          -56,-32     , -179,-107   , -369,-193 ,  &
          -56,-31     , -101,-53    , -156,-74  ,  &
          -34,-22     , -67,-22     , -89,0     ,  &
          0,10        , -23,53      , -68,128   ,  &
          -55,107     , -78,171     , -89,214   ,  &
          -11,54      , 34,118      , 134,182   ,  &
          0,0         , 67,32       , 212,107   ,  &
          45,22       , 123,53      , 235,96    ,  & 
          100,33      , 178,65      , 223,86    ,  & 
          23,11       , 23,43       , 0,96 ],  &
        [2, 3, 48 ] ) / 100.0_FLOAT
    
    REAL, PARAMETER :: g_curves(2,3,34) = RESHAPE(  &
        [ -23,43      , -45,53      , -67,43    ,  &
          -22,-11     , -67,-54     , -134,-118 ,  &
          -79,-75     , -123,-118   , -134,-139 ,  &
          -257,32     , -481,-11    , -670,-96  ,  &
          -302,-139   , -469,-407   , -492,-792 ,  &
          -33,-375    , 56,-792     , 280,-1231 ,  &
          234,-450    , 524,-771    , 859,-985  ,  &
          347,-225    , 682,-267    , 983,-118  ,  &
          134,65      , 257,182     , 335,321   ,  &
          90,161      , 134,343     , 123,546   ,  &
          0,43        , 0,75        , -11,86    ,  &
          -22,43      , -123,128    , -279,246  ,  &
          -168,118    , -268,161    , -291,150  ,  &
          -11,0       , -11,-43     , -22,-118  ,  &
          -11,-96     , -33,-171    , -56,-224  ,  &
          -44,-97     , -111,-172   , -201,-215 ,  &
          -145,-64    , -301,-43    , -469,86   ,  &
          -156,107    , -290,268    , -402,492  ,  &
          -123,225    , -179,429    , -168,610  ,  &
          12,172      , 101,300     , 257,375   ,  &
          90,54       , 224,64      , 380,54    ,  &
          22,-54      , 78,-183     , 156,-375  ,  &
          68,-150     , 123,-278    , 179,-364  ,  &
          11,-32      , 22,-43      , 34,-43    ,  &
          11,0        , 22,0        , 33,0      ,  &
          23,21       , 67,43       , 123,75    ,  &
          45,32       , 89,64       , 112,75    ,  &
          33,11       , 78,32       , 145,54    ,  &
          67,21       , 112,42      , 145,64    ,  &
          23,10       , 23,32       , 0,85      ,  &
          -33,64      , -89,161     , -167,300  ,  &
          -79,139     , -134,236    , -168,300  ,  &
          -100,182    , -235,460    , -402,834  ,  &
          0,11        , -11,11      , -11,22 ],  &
        [2, 3, 34] ) / 100.0_FLOAT
    
    REAL, PARAMETER :: m2_curves(2,3,44) = RESHAPE(  &
        [ -10,32      , -44,43      , -66,43    ,  &
          -179,-65    , -358,-139   , -536,-204 ,  &
          -23,-10     , -23,-32     , 0,-85     ,  &
          66,-129     , 167,-321    , 312,-578  ,  &
          146,-257    , 246,-450    , 313,-578  ,  &
          22,-43      , 34,-64      , 22,-64    ,  &
          -11,0       , -189,160    , -502,513  ,  &
          -201,204    , -391,407    , -581,610  ,  &
          -22,22      , -45,33      , -67,33    ,  &
          0,0         , -22,0       , -56,-22   ,  &
          -56,-11     , -78,-43     , -78,-75   ,  &
          22,-160     , 67,-406     , 134,-728  ,  &
          67,-342     , 100,-588    , 134,-738  ,  &
          0,-11       , 0,-22       , 0,-32     ,  &
          -11,10      , -34,21      , -44,53    ,  &
          -68,129     , -168,332    , -302,600  ,  &
          -134,267    , -246,470    , -313,610  ,  &
          -22,42      , -45,64      , -78,53    ,  &
          -157,-64    , -324,-139   , -492,-203 ,  &
          -22,-11     , -22,-21     , -22,-32   ,  &
          0,-11       , 0,-21       , 11,-54    ,  &
          79,-149     , 201,-374    , 380,-674  ,  &
          167,-300    , 290,-524    , 369,-674  ,  &
          78,-161     , 234,-461    , 457,-910  ,  &
          56,-118     , 146,-289    , 257,-525  ,  &
          0,-10       , 12,-21      , 12,-32    ,  &
          11,-32      , 44,-32      , 78,-11    ,  &
          56,33       , 134,65      , 234,119   ,  &
          112,53      , 190,96      , 235,117   ,  &
          11,11       , 11,43       , 0,118     ,  &
          -89,503     , -179,1006   , -268,1498 ,  &
          -11,107     , -11,161     , 11,172    ,  &
          11,0        , 56,-33      , 145,-118  ,  &
          235,-236    , 592,-568    , 1061,-1017,  &
          45,-43      , 78,-54      , 90,-43    ,  &
          33,11       , 66,43       , 123,75    ,  &
          44,32       , 89,64       , 111,75    ,  &
          22,11       , 67,32       , 112,54    ,  &
          44,21       , 89,42       , 112,53    ,  &
          22,22       , 33,43       , 11,75     ,  &
          -79,139     , -190,364    , -346,642  ,  &
          -168,289    , -280,503    , -358,653  ,  &
          -78,139     , -179,343    , -301,610  ,  &
          -135,278    , -235,482    , -314,621 ],  &
        [2, 3, 44] ) / 100.0_FLOAT
    
    REAL, PARAMETER :: s_curves(2,3,25) = RESHAPE(  &
        [ -100,214    , -290,353    , -536,459  ,  &
          -223,86     , -458,108    , -692,86   ,  &
          -11,0       , -23,-10     , -23,-10   ,  &
          -22,0       , -33,-108    , -22,-290  ,  &
          0,-192      , 11,-310     , 34,-342   ,  &
          11,-11      , 22,-21      , 33,-21    ,  &
          168,21      , 279,21      , 335,21    ,  &
          134,-21     , 224,-85     , 291,-203  ,  &
          33,-64      , 44,-161     , 22,-289   ,  &
          -34,-203    , -45,-322    , -45,-343  ,  &
          0,-150      , 34,-299     , 112,-449  ,  &
          111,-214    , 279,-343    , 503,-385  ,  &
          178,-43     , 368,-11     , 558,74    ,  &
          11,11       , 23,11       , 23,22     ,  &
          21,0        , 21,21       , 11,53     ,  &
          -11,96      , -34,268     , -56,525   ,  &
          0,0         , 0,10        , 0,10      ,  &
          -12,22      , -23,22      , -34,11    ,  &
          -11,0       , -22,-11     , -33,-21   ,  &
          -23,-11     , -34,-22     , -45,-33   ,  &
          -67,-31     , -134,-53    , -201,-42  ,  &
          -78,11      , -134,42     , -179,117  ,  &
          -33,76      , -44,172     , -22,300   ,  &
          22,236      , 44,354      , 44,364    ,  &
          12,139      , -11,278     , -78,386 ],  &
        [2,3,25] ) / 100.0_FLOAT
    
    !***************************************************************************
    
    !---------------------------------------------------------------------------
    ! Upper seam shape.
    
    CALL CreatePathGeometry(  &
        factory,  &
        test%upper_seam,  &
        hr)
    IF (hr < 0) RETURN
    CALL Open(test%upper_seam, sink, hr)
    IF (hr < 0) RETURN
    
    point = D2D1_POINT_2F(-31.45_FLOAT, 47.11_FLOAT)
    CALL BeginFigure(sink, point, D2D1_FIGURE_BEGIN_FILLED)
    point = point + D2D1_SIZE_F(-4.69_FLOAT, 14.88_FLOAT)
    CALL AddLine(sink, point)
    point = point + D2D1_SIZE_F(251.86_FLOAT, 73.20_FLOAT)
    CALL AddLine(sink, point)
    point = point + D2D1_SIZE_F(4.69_FLOAT, -14.98_FLOAT)
    CALL AddLine(sink, point)
    CALL EndFigure(sink, D2D1_FIGURE_END_CLOSED)
    
    CALL Close(sink, hr)
    IF (hr < 0) RETURN
    
    !---------------------------------------------------------------------------
    ! Lower seam shape.
    
    CALL CreatePathGeometry(  &
        factory,  &
        test%lower_seam,  &
        hr)
    IF (hr < 0) RETURN
    CALL Open(test%lower_seam, sink, hr)
    IF (hr < 0) RETURN
    
    CALL BeginFigure(  &
        sink,  &
        D2D1_POINT_2F(0.00_FLOAT, 85.85_FLOAT),  &
        D2D1_FIGURE_BEGIN_FILLED )
    CALL AddLines(  &
        sink,  &
        [ D2D1_POINT_2F(139.21_FLOAT, 133.16_FLOAT),  &
          D2D1_POINT_2F(113.41_FLOAT, 134.98_FLOAT) ] )
    CALL EndFigure(sink, D2D1_FIGURE_END_CLOSED)
    
    CALL Close(sink, hr)
    IF (hr < 0) RETURN
    
    !---------------------------------------------------------------------------
    ! The top right ellipse.
    
    CALL CreateEllipseGeometry(  &
        factory,  &
        D2D1_ELLIPSE( &
          !D2D1_POINT_2F(155.25_FLOAT, 22.01_FLOAT),  &
          D2D1_POINT_2F(0.0_FLOAT, 0.0_FLOAT),  &
          132.00_FLOAT,  &
          45.00_FLOAT ),  &
        ellipse,  &
        hr )
    IF (hr < 0) RETURN
    CALL CreateTransformedGeometry(  &
        factory,  &
        ellipse,  &
        Rotation(23.75) * Translation(155.25_FLOAT, 22.01_FLOAT),  &
        test%ellipse,  &
        hr )
    IF (hr < 0) RETURN
    
    !---------------------------------------------------------------------------
    ! The letters.
    
    CALL CreatePathGeometry(  &
        factory,  &
        test%letters,  &
        hr)
    IF (hr < 0) RETURN
    CALL Open(test%letters, sink, hr)
    IF (hr < 0) RETURN
    
    ! M
    CALL add_letter_curves_to_sink(  &
        sink,  &
        D2D1_POINT_2F(90.96_FLOAT, 35.12_FLOAT),  &
        m1_curves )
    
    !.
    CALL add_circle_to_sink(  &
        sink,  &
        D2D1_POINT_2F(93.75_FLOAT, 36.62_FLOAT),  &
        1.60_FLOAT )
    
    ! E
    CALL add_letter_curves_to_sink(  &
        sink,  &
        D2D1_POINT_2F(115.53_FLOAT, 41.11_FLOAT),  &
        e_curves )
    
    !.
    CALL add_circle_to_sink(  &
        sink,  &
        D2D1_POINT_2F(114.86_FLOAT, 49.14_FLOAT),  &
        1.60_FLOAT )
    
    ! G
    CALL add_letter_curves_to_sink(  &
        sink,  &
        D2D1_POINT_2F(136.53_FLOAT, 58.56_FLOAT),  &
        g_curves )
    
    !.
    CALL add_circle_to_sink(  &
        sink,  &
        D2D1_POINT_2F(139.21_FLOAT, 60.05_FLOAT),  &
        1.60_FLOAT )
    
    ! M
    CALL add_letter_curves_to_sink(  &
        sink,  &
        D2D1_POINT_2F(164.89_FLOAT, 70.12_FLOAT),  &
        m2_curves )
    
    !.
    CALL add_circle_to_sink(  &
        sink,  &
        D2D1_POINT_2F(168.47_FLOAT, 70.86_FLOAT),  &
        1.60_FLOAT )
    
    ! S
    CALL add_letter_curves_to_sink(  &
        sink,  &
        D2D1_POINT_2F(186.56_FLOAT, 69.59_FLOAT),  &
        s_curves )
    
    !.
    CALL add_circle_to_sink(  &
        sink,  &
        D2D1_POINT_2F(184.67_FLOAT, 77.08_FLOAT),  &
        1.60_FLOAT )
    
    CALL Close(sink, hr)
    IF (hr < 0) RETURN
    
  END SUBROUTINE construct_logo_test
  
  
  !*****************************************************************************
  !!
  !> Make a figure from the SVG "c" style relative bezier curve specification.
  
  SUBROUTINE add_letter_curves_to_sink(sink, initial_point, curves)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The geometry sink to add the curves to.
    TYPE(ID2D1GeometrySink), INTENT(IN) :: sink
    !> The initial point of the curve.
    TYPE(D2D1_POINT_2F), INTENT(IN) :: initial_point
    !> The bezier control points and intermediate points of the curves.
    REAL(FLOAT), INTENT(IN) :: curves(:,:,:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: icurve         ! Curve index.
    
    ! Current point.
    TYPE(D2D1_POINT_2F) :: point
    
    !***************************************************************************
    
    point = initial_point
    CALL BeginFigure(sink, point, D2D1_FIGURE_BEGIN_FILLED)
    DO icurve = 1, SIZE(curves, 3)
      CALL AddBezier(  &
          sink,  &
          D2D1_BEZIER_SEGMENT(  &
            point + D2D1_SIZE_F(curves(1,1,icurve), curves(2,1,icurve)),  &
            point + D2D1_SIZE_F(curves(1,2,icurve), curves(2,2,icurve)),  &
            point + D2D1_SIZE_F(curves(1,3,icurve), curves(2,3,icurve)) ) )
      point = point + D2D1_SIZE_F(curves(1,3,icurve), curves(2,3,icurve))
    END DO
    CALL EndFigure(sink, D2D1_FIGURE_END_CLOSED)
    
  END SUBROUTINE add_letter_curves_to_sink
  
  
  !*****************************************************************************
  !!
  !> Add a circular shape to a geometry sink.
  
  SUBROUTINE add_circle_to_sink(sink, centre_point, radius)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The geometry sink to add the curves to.
    TYPE(ID2D1GeometrySink), INTENT(IN) :: sink
    !> The centre point of the circle.
    TYPE(D2D1_POINT_2F), INTENT(IN) :: centre_point
    !> The radius of the circle.
    REAL(FLOAT), INTENT(IN) :: radius
    
    !***************************************************************************
    
    CALL BeginFigure(  &
        sink,  &
        centre_point + D2D1_SIZE_F(-radius, 0.0_FLOAT),  &
        D2D1_FIGURE_BEGIN_FILLED )
    CALL AddArc(  &
        sink,  &
        D2D1_ARC_SEGMENT(  &
          centre_point + D2D1_SIZE_F(radius, 0.0_FLOAT),  &
          D2D1_SIZE_F(radius, radius),  &
          180.0_FLOAT,  &
          D2D1_SWEEP_DIRECTION_COUNTER_CLOCKWISE,  &
          D2D1_ARC_SIZE_LARGE ) )
    CALL AddArc(  &
        sink,  &
        D2D1_ARC_SEGMENT(  &
          centre_point + D2D1_SIZE_F(-radius, 0.0_FLOAT),  &
          D2D1_SIZE_F(radius, radius),  &
          180.0_FLOAT,  &
          D2D1_SWEEP_DIRECTION_COUNTER_CLOCKWISE,  &
          D2D1_ARC_SIZE_LARGE ) )
    CALL EndFigure(sink, D2D1_FIGURE_END_CLOSED)
    
  END SUBROUTINE add_circle_to_sink
  
  
  !*****************************************************************************
  !!
  !> Create the device dependent resources for the logo test.
  
  SUBROUTINE logo_CreateResources(test, render_target, hr)
    
    USE Direct2D
    USE Direct2DColor
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(logo_test), INTENT(INOUT) :: test
    !> The current render target.
    CLASS(ID2D1RenderTarget), INTENT(IN) :: render_target
    !> COM result code - negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !***************************************************************************
    
    CALL CreateSolidColorBrush(  &
        render_target,  &
        D2D1_COLOR_F(  &
          204 / 255.0_FLOAT,  &
          102 / 255.0_FLOAT,  &
          0 / 255.0_FLOAT ),  &
        test%waste,  &
        hr )
    
    CALL CreateSolidColorBrush(  &
        render_target,  &
        Black,  &
        test%coal,  &
        hr )
    
    CALL CreateSolidColorBrush(  &
        render_target,  &
        White,  &
        test%hole,  &
        hr )
    
  END SUBROUTINE logo_CreateResources
  
  
  !*****************************************************************************
  !!
  !> Discard the device dependent resources for the logo test.
  
  SUBROUTINE logo_DiscardResources(test)
    
    USE COMBase
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(logo_test), INTENT(INOUT) :: test
    
    !***************************************************************************
    
    IF (Exists(test%waste))             CALL Release(test%waste)
    IF (Exists(test%coal))              CALL Release(test%coal)
    IF (Exists(test%hole))              CALL Release(test%hole)
    
  END SUBROUTINE logo_DiscardResources
  
  
  !*****************************************************************************
  !!
  !> Draw a logo, that was initially defined by a SVG file.
  
  SUBROUTINE logo_Draw(test, render_target, hr)
    
    USE Direct2D
    USE DirectWrite
    USE Direct2DColor
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    !> The test object.
    CLASS(logo_test), INTENT(INOUT) :: test
    !> The current render target.
    CLASS(ID2D1RenderTarget), INTENT(IN) :: render_target
    !> COM result code - negative on failure.
    INTEGER(HRESULT), INTENT(OUT) :: hr
    
    !***************************************************************************
    
    ! For now, map the SVG view box of "0 0 215.90 279.40" across to 
    ! the window extents.
    BLOCK
      ! Render target size.
      TYPE(D2D1_SIZE_F) :: size
      ! The scale factor
      REAL(FLOAT) :: factor
      ! The transform.
      TYPE(D2D1_MATRIX_3X2_F) :: txfrm
      
      size = GetSize(render_target)
      factor = MIN(size.width / 194.00_FLOAT, size.height / 125.40_FLOAT)
      
      txfrm = Scale(factor, factor, D2D1_POINT_2F())  &
          * Translation(  &
            size.width / 2 - factor * 194.00_FLOAT / 2,  &
            size.height / 2 - factor * 125.00_FLOAT / 2 )
      
      CALL SetTransform(  &
          render_target,  &
          txfrm )
    END BLOCK
    
    CALL BeginDraw(render_target)
    CALL Clear(render_target, White)
    
    CALL PushAxisAlignedClip(  &
        render_target,  &
        D2D1_RECT_F_IfortBug( &
          D2D1_POINT_2F(0.0_FLOAT, 0.0_FLOAT),  &
          D2D1_SIZE_F(194.00_FLOAT, 125.00_FLOAT) ),  &
        D2D1_ANTIALIAS_MODE_PER_PRIMITIVE )
    
    CALL FillRectangle(  &
        render_target,  &
        D2D1_RECT_F_IfortBug(  &
          D2D1_POINT_2F(0.00_FLOAT, 0.00_FLOAT),  &
          D2D1_SIZE_F(194.00_FLOAT, 125.00_FLOAT) ),  &
        test%waste )
    CALL FillRectangle(  &
        render_target,  &
        D2D1_RECT_F_IfortBug(  &
          D2D1_POINT_2F(34.00_FLOAT, -0.50_FLOAT),  &
          D2D1_SIZE_F(10.00_FLOAT, 72.00_FLOAT) ),  &
        test%hole )
    CALL FillGeometry(  &
        render_target,  &
        test%upper_seam,  &
        test%coal )
    CALL FillGeometry(  &
        render_target,  &
        test%lower_seam,  &
        test%coal )
    CALL FillGeometry(  &
        render_target,  &
        test%ellipse,  &
        test%hole )
    
    CALL FillGeometry(  &
        render_target,  &
        test%letters,  &
        test%coal )
    
    CALL PopAxisAlignedClip(render_target)
    
    CALL EndDraw(render_target, hr)
    
  END SUBROUTINE logo_Draw
  
  
  ! Finalization of components of logo_test should happen automatically, 
  ! but under certain circumstances (e.g start the program with the logo 
  ! test, then immediately exit it) some of the final procedures are not 
  ! being invoked, leading to Direct2D complaining about a resource 
  ! leak.
  !
  ! (The one instance of logo test should be deallocated and hence 
  ! finalized by the explicit deallocate of the current test at 
  ! the end of the WinMain procedure.)
  !
  ! I suspect this is an ifort bug.  Haven't cut out a reproducer yet.
  subroutine logo_final_suspect_ifort_bug(obj)
    use combase
    type(logo_test), intent(inout) :: obj
    
    call release(obj%letters)
    call release(obj%ellipse)
    call release(obj%lower_seam)
    call release(obj%upper_seam)
  end subroutine logo_final_suspect_ifort_bug
  
END MODULE Direct2DTest
